import os
from pesto.ws.features.converter.array_converter import ArrayConverter
from pesto.ws.features.converter.metadata_converter import MetadataConverter
from pesto.ws.features.converter.object_converter import ObjectConverter
from pesto.ws.features.converter.primitive_converter import PrimitiveConverter
from pesto.ws.features.converter.image_converter import ImageConverter

primitive_schemas = [{'type': _} for _ in ['number', 'string', 'float', 'int', 'bool']]
metadata_schemas = [{'$ref': _} for _ in ['#/definitions/Metadata', '#/definitions/Metadatas']]
image_schemas = [{'$ref': _} for _ in ['#/definitions/Image', '#/definitions/Images']]


def convert(x, schema):
    return 1 if x == 2 else 0


def test_image_converter():
    c = ImageConverter(None)

    value = os.path.join(os.path.dirname(__file__), "resources", "image.tif")
    value = 'file://{}'.format(value)

    for schema in image_schemas:
        assert c.match(schema)
        output = c.convert((value, schema))
        assert len(output.shape) == 3
    for schema in primitive_schemas:
        assert c.match(schema) is False


def test_metadata_converter():
    c = MetadataConverter()
    value = 'this is a random value'

    for schema in metadata_schemas:
        assert c.match(schema)
        assert c.convert((value, schema)) == value
    for schema in primitive_schemas:
        assert c.match(schema) is False


def test_primitive_converter():
    c = PrimitiveConverter()
    value = 'this is a random value'
    for schema in primitive_schemas:
        assert c.match(schema)
        assert c.convert((value, schema)) == value


def test_array_converter():
    c = ArrayConverter(convert)
    value = {'items': [2, 2, 1, 2, 3]}
    schema = {
        'type': 'array',
        'items': 'number'
    }

    assert c.match(schema)
    output = c.convert((value, schema))
    assert output == [convert(_, None) for _ in value]


def test_object_converter():
    c = ObjectConverter(convert)
    value = {
        'a': 2,
        'b': 'toto'
    }
    schema = {
        'type': 'object',
        'properties': {
            'a': {
                'type': 'number'
            },
            'b': {
                'type': 'string'
            }
        }
    }
    assert c.match(schema)

    output = c.convert((value, schema))
    for k, v in output.items():
        assert v == convert(value[k], schema['properties'][k])



import time
from PIL import Image
import numpy as np


def create_box(x, y, width, height):
    return {
        "type": "Polygon",
        "coordinates": [[[x, y], [x, y + height], [x + width, y + height], [x + width, y], [x, y]]],
    }


class Process:
    """
    Custom processing class
    """
    MODEL = None

    # Do not call __init__

    def on_start(self) -> None:
        """
        Process.on_start will be called at server start time.
        If you need to load heavy resources before processing data, this should be done here.
        """
        def my_model(image: np.ndarray, gamma: float):
            # Image is parsed as (C, H, W) - rasterio convention, so we put it back to (H, W C)
            img = np.transpose(image, (1, 2, 0))

            # A dummy model that modifies gamma
            img = img.astype(np.float32) / 255.0
            img = img**gamma
            img = np.clip(img, 0.0, 1.0)

            img = (255.0 * img).astype(np.uint8)

            # Put image back to (H, W, C) format for response parsing
            img = np.transpose(img, (2, 0, 1))

            return img

        Process.MODEL = staticmethod(my_model)

    # process function, called each time a processing request is send to the service.
    # note: images are represented in numpy as follows [channel, hight, width].
    def process(
        self,
        image,
        dict_parameter,
        object_parameter,
        integer_parameter,
        number_parameter,
        string_parameter,
    ):
        """
        The core algorithm is implemented here.
        """

        # Defaults
        dict_parameter = dict_parameter or dict()
        object_parameter = object_parameter or dict()
        integer_parameter = integer_parameter or 0
        number_parameter = number_parameter or 0.7
        string_parameter = string_parameter or "defaultString"

        dict_parameter["object"] = object_parameter

        # Simulates longer algorithm
        time.sleep(5)

        h, w = image.shape[1], image.shape[2]

        # Apply a model
        output = self.MODEL(image=image, gamma=number_parameter)

        # Create a list of polygons for demo purposes
        areas = [
            create_box(0, 0, width=w, height=h),
            create_box(int(w // 4), int(h // 4), width=int(w // 2), height=int(h // 2)),
        ]

        # Create a list of image for demo purposes
        image_list = [
            self.MODEL(image=image, gamma=0.7),
            self.MODEL(image=image, gamma=1.5),
        ]

        result = {
            "image": output,
            "number_output": number_parameter,
            "integer_output": integer_parameter,
            "string_output": string_parameter,
            "dict_output": dict_parameter,
            "areas": areas,
            "image_list": image_list,
            "geojson": {
                "features": [
                    {
                        "geometry": create_box(0, 0, width=w, height=h),
                        "properties": {
                            "category": "cat",
                            "confidence": 0.99,
                            "name": "chelsea"
                        },
                        "type": "Feature",
                    },
                    {
                        "geometry": create_box(int(w // 4), int(h // 4), width=int(w // 2), height=int(h // 2)),
                        "properties": {
                            "category": "egyptian_cat",
                            "confidence": 0.42,
                        },
                        "type": "Feature",
                    },
                ]
            },
        }

        return result



import os

import numpy as np
import rasterio
from PIL import Image as PILImage
from pesto.ws.features.converter.image.image import Image

# os.environ['REQUESTS_CA_BUNDLE'] = '/etc/ssl/certs/ca-certificates.crt'
print('rasterio version=' + rasterio.__version__)
print('gdal version=' + rasterio.gdal_version())


def open_url(url: str):
    with rasterio.Env(GDAL_HTTP_UNSAFESSL='YES'):
        image = Image.from_uri(url).to_array()
    return image
    # try:
    #     with rasterio.open(url) as _:
    #         return _.read()
    # except:
    #     # log.debug('could not open : {}'.format(url))
    #     pass
    #
    # try:
    #     url_stream = '/vsicurl_streaming/' + url
    #     # log.debug('trying with streaming: {}'.format(url_stream))
    #     with rasterio.open(url_stream) as _:
    #         return _.read()
    # except:
    #     raise ValueError('Could not open image: {}'.format(url))


def test_sobloo_url():
    url = "https://sobloo.eu/api/v1/services/wmts/4004658d-f245-4420-8993-ebc49f80cc37/tiles/1.0.0/default/rgb/EPSG4326/8/259/64.png"
    rio_image = open_url(url)

    png_image = PILImage.open(os.path.join(os.path.dirname(__file__), "resources", "sobloo.png"))
    png_image = np.asarray(png_image)

    assert np.all(rio_image == png_image.transpose((2, 0, 1)))



import collections as collections_abc
import os
from typing import List, Any, Dict, Union

import six

from pesto.common.utils import load_json


class ProfileLoader(object):
    def __init__(self, root_path: str, profiles: List[str]):
        self.root_path = root_path
        self.profiles = profiles

    def load(self, filename: str) -> Any:
        full_path = os.path.join(self.root_path, filename)
        content = load_json(full_path)
        for profile in self.profiles:
            try:
                profile_content = load_json(profile_path(profile, full_path))
                content = _update(content, profile_content)
            except:
                pass
        return content


def profile_path(profile: str, path: str) -> Union[str, None]:
    new_path = '.{}'.format(profile).join(os.path.splitext(path))
    if profile is not None and os.path.exists(new_path):
        return new_path
    return None


def _update(d: Dict, u: Dict) -> Dict:
    for k, v in six.iteritems(u):
        dv = d.get(k, {})
        if not isinstance(dv, collections_abc.Mapping):
            d[k] = v
        elif isinstance(v, collections_abc.Mapping):
            d[k] = _update(dv, v)
        else:
            d[k] = v
    return d



from typing import Any, Tuple

from pesto.ws.core.match_apply import MatchApply, Json


class PrimitiveConverter(MatchApply):
    def match(self, schema: Json):
        return schema.get("type") is not None

    def convert(self, data: Tuple[Any, Json]):
        payload, schema = data
        return payload



from typing import Tuple

import numpy as np


class BandInterface(object):
    def compute_crop(self, image: np.ndarray) -> Tuple[int, int]:
        raise NotImplementedError()


class Band(BandInterface):
    def __init__(self, config: dict):
        self.first = config['first']
        self.number = config['number']
        self.margin = config.get('margin', 0)
        self.last = self.first + self.number
        if self.margin < 0:
            raise ValueError('The margin must be greather than 0 : {}'.format(self.margin))

    def _check_range(self, size: int) -> None:
        last = self.first + self.number
        if not (self.first >= 0 and self.number > 0 and last <= size):
            raise ValueError('[{},{}[ are out of range for a size of {}'.format(self.first, last, size))

    def compute_crop(self, size: int) -> Tuple[int, int]:
        self._check_range(size)

        first_with_margin = max(0, self.first - self.margin)
        last_with_margin = min(size, self.last + self.margin)
        before = self.first - first_with_margin
        after = last_with_margin - self.last

        return before, after


class FullBand(BandInterface):
    def compute_crop(self, image: np.ndarray) -> Tuple[int, int]:
        return 0, 0



import json
import time
from tempfile import NamedTemporaryFile

import requests
from pesto.common.testing import logger
from pesto.common.utils import truncate_dict_for_debug
from pesto.ws.features.converter.image.image import Image


class EndpointManager:
    def __init__(self, server_url):
        self.server_url = server_url
        self._describe = None

    @property
    def is_alive(self) -> bool:
        try:
            response = requests.get("{}/api/v1/health".format(self.server_url))

            return response.status_code == 200
        except:
            return False

    @property
    def describe(self) -> dict:
        if not self._describe:
            self._describe = requests.get(url="{}/api/v1/describe".format(self.server_url)).json()
        return self._describe

    @property
    def stateful(self) -> bool:
        return self.describe.get("asynchronous")

    @property
    def input_schema(self) -> dict:
        return self.describe.get("input")

    @property
    def output_schema(self) -> dict:
        return self.describe.get("output")

    def process(self, payload: dict) -> dict:
        logger.debug(json.dumps(truncate_dict_for_debug(payload), indent=2))
        response = requests.post(url="{}/api/v1/process".format(self.server_url), json=payload)

        if self.stateful:
            status_url = response.json()["link"]
            job_id = status_url.split("/")[-2]
            job = _StatefulJob(
                server_url=self.server_url,
                job_id=job_id,
                output_schema=self.output_schema,
            )
            if job.wait_done():
                result = job.result
            else:
                result = {"status": "PROCESSING ERROR"}
        else:
            job = _StatelessJob(response=response)
            result = job.result

        logger.debug(json.dumps(truncate_dict_for_debug(result), indent=2))
        return result


class _StatelessJob:
    """
    Internal class to parse results from a stateless job
    """

    def __init__(self, response):
        self.response = response

    @property
    def result(self):
        return self.response.json()


class _StatefulJob:
    """
    Internal class to parse results from a stateful job
    """

    def __init__(self, server_url, job_id, output_schema):
        self.server_url = server_url
        self.job_id = job_id
        self.output_schema = output_schema

    @property
    def status_url(self):
        return "{}/api/v1/jobs/{}/status".format(self.server_url, self.job_id)

    @property
    def results_url(self):
        return "{}/api/v1/jobs/{}/results".format(self.server_url, self.job_id)

    @property
    def is_done(self):
        status = requests.get(self.status_url).json().get("status")
        return status == "DONE"

    def wait_done(self, max_tries=20):
        num_tries = 0
        while (num_tries < max_tries) and (not self.is_done):
            logger.info("Waiting for job to complete")
            num_tries += 1
            time.sleep(2)

        if num_tries < max_tries:
            logger.info("Job complete")
            return True
        else:
            logger.error("Timeout reached")
            return False

    @property
    def result(self):
        result = requests.get(self.results_url).json()
        result = self._parse_result(result)
        return result

    def _parse_result(self, result: dict):
        parsed_result = dict()
        for key in result:
            link_type = self._compute_type(key)
            val = result.get(key)
            if self._is_uri(val):
                if isinstance(val, list):
                    parsed_result[key] = [self._get_result(uri, link_type) for uri in val]
                else:
                    parsed_result[key] = self._get_result(val, link_type)
            else:
                parsed_result[key] = val

        return parsed_result

    @staticmethod
    def _get_result(uri, key_type):
        response = requests.get(uri)
        content_type = response.headers["Content-Type"]

        if key_type == "#/definitions/Images":
            driver = content_type.split("/")[-1]
            path = "{}.{}".format(NamedTemporaryFile().name, driver)
            result = Image.from_bytes(response.content).to_path(path)
        elif key_type == "#/definitions/Image":
            driver = content_type.split("/")[-1]
            path = "{}.{}".format(NamedTemporaryFile().name, driver)
            result = Image.from_bytes(response.content).to_path(path)
        elif key_type in [
            "#/definitions/Metadata",
            "#/definitions/Metadatas",
            "#/definitions/Polygon",
            "#/definitions/Polygons",
            "#/definitions/Tag",
            "#/definitions/Tags",
        ]:
            result = response.json()
        elif key_type == "number":
            result = response.json()
        elif key_type == "string":
            result = response.text
        else:
            result = response.json()

        return result

    def _is_uri(self, val: [str, list]):
        if isinstance(val, str):
            return val.startswith(self.server_url)
        else:
            return all([self._is_uri(v) for v in val])

    def _compute_type(self, key: str):

        schema = self.output_schema.get("properties")

        key_type = schema.get(key).get("$ref") or schema.get(key).get("type")

        return key_type



import logging
from typing import Any

from pesto.ws.core.pesto_feature import PestoFeature
from pesto.ws.service.job_result import ResultType

log = logging.getLogger(__name__)


class StatefulResponse(PestoFeature):

    def __init__(self, url_root: str, job_id: str):
        self.url_root = url_root
        self.job_id = job_id

    def process(self, payload: dict) -> Any:
        log.info('response mode : stateful')
        return {
                   'link': '{}/api/v1/jobs/{}/status'.format(self.url_root, self.job_id)
               }, ResultType.json



import logging
import os
from typing import Tuple, Any

from pesto.common.utils import save_json
from pesto.ws.core.match_apply import MatchApply, Json

log = logging.getLogger(__name__)


class JsonSerializer(MatchApply):
    def __init__(self, job_path: str):
        self.job_path = job_path

    def match(self, schema: Json):
        ref = schema.get('$ref')
        return ref in [
            '#/definitions/Metadata',
            '#/definitions/Metadatas',
            '#/definitions/Polygon',
            '#/definitions/Polygons'
        ]

    def convert(self, data: Tuple[Any, str]):
        payload, key = data
        filename = os.path.join(self.job_path, key + '.json')
        log.info('saving json : ' + filename)
        save_json(filename, payload)



from typing import Callable, Any, Tuple

from pesto.ws.core.match_apply import MatchApply, Json


class ArrayConverter(MatchApply):

    def __init__(self, convert: Callable):
        self.delegate = convert

    def match(self, schema: Json):
        return schema.get('type') == 'array'

    def convert(self, data: Tuple[Any, Json]):
        payload, schema = data
        if payload is None:
            return payload
        return [self.delegate(e, schema['items']) for e in payload]



import logging
from enum import Enum
from typing import Dict

from pesto.ws.core.pesto_feature import PestoFeature
from pesto.ws.features.converter.image.image_roi import ImageROI

log = logging.getLogger(__name__)


class PestoConfig(Enum):
    __order__ = ['roi']
    roi = ImageROI

    def build(self, param):
        return self.value(param)


class PayloadParser(object):
    PIPELINE = ['roi']
    FEATURES = {
        'roi': ImageROI
    }

    @staticmethod
    def parse(payload: dict) -> Dict[PestoConfig, PestoFeature]:
        log.info('pesto features extraction ...')

        config = payload.pop('pesto') if 'pesto' in payload else {}

        features = {
            item: item.build(config[item.name])
            for item in PestoConfig if item.name in config
        }
        for name in filter(lambda _: _ in config, PayloadParser.PIPELINE):
            features[name] = PayloadParser.FEATURES[name](config[name])
        return features



import logging

from pesto.ws.core.pesto_feature import PestoFeature

log = logging.getLogger(__name__)


class PayloadDebug(PestoFeature):
    def __init__(self, schema: dict):
        self.schema = schema

    def process(self, payload: dict) -> dict:
        log.info('listing inputs')
        for key, value in self.schema['properties'].items():
            key_type = value.get('$ref', value)
            try:
                key_infos = payload[key].shape
            except:
                key_infos = 'json'
            log.info('{} : {} : {}'.format(key, key_type, key_infos))
        return payload



import logging

import jsonschema

from pesto.ws.core.pesto_feature import PestoFeature

log = logging.getLogger(__name__)


class SchemaValidation(PestoFeature):
    def __init__(self, schema: dict):
        self.schema = schema

    def process(self, payload: dict) -> dict:
        jsonschema.validate(payload, self.schema)
        return payload



import json
import os
from pathlib import Path
from typing import List
from urllib.parse import urlparse

from pesto.common.pesto import PESTO_WORKSPACE
from pesto.common.utils import load_json


class BuildConfig:

    @staticmethod
    def from_path(path: str,
                  profiles: List[str] = None,
                  proxy: str = None,
                  pip_extra_index: str = None,
                  network: str = None
                  ):
        assert path is not None



        build_config = load_json(path)
        return BuildConfig(
            name=build_config['name'],
            version=build_config['version'],
            profiles=profiles,
            workspace=build_config.get('workspace'),
            algorithm_path=build_config.get('algorithm_path') or str(Path(path).parent.parent.parent),
            proxy=proxy,
            pip_extra_index=pip_extra_index,
            network=network)

    def __init__(self,
                 name: str = None,
                 version: str = None,
                 profiles: List[str] = None,
                 workspace: str = None,
                 algorithm_path: str = None,
                 proxy: str = None,
                 pip_extra_index: str = None,
                 network: str = None):
        self.name = name
        self.version = version
        self.algorithm_path = algorithm_path

        self.profiles = profiles or []
        self.proxy = proxy or ''
        self.network = network
        self.pip_extra_index = pip_extra_index or os.environ.get('PIP_EXTRA_INDEX_URL')
        self.workspace = workspace or os.path.join(PESTO_WORKSPACE, self.name, self.full_version)

    @property
    def full_version(self):
        if self.profiles is not None and len(self.profiles) > 0:
            profile_string = '-'.join(self.profiles)
            return '{}-{}'.format(self.version, profile_string)
        return self.version

    @property
    def docker_image_name(self):
        return '{}:{}'.format(self.name, self.full_version)

    @property
    def pip_proxies(self):
        proxies = ['pypi.python.org', 'pypi.org', 'files.pythonhosted.org']

        pip_extra_index = self.pip_extra_index
        if pip_extra_index:
            parsed_uri = urlparse(pip_extra_index)
            parsed_uri = parsed_uri.netloc
            proxies.append(parsed_uri)

        pip_string = ' '.join(['--trusted-host={}'.format(_) for _ in proxies])
        index_string = ' -i {}'.format(pip_extra_index) if pip_extra_index else ''
        return '{}{}'.format(pip_string, index_string)

    def to_dict(self) -> dict:
        return {
            'name': self.name,
            'version': self.full_version,
            'workspace': self.workspace,
            'algorithm_path': self.algorithm_path
        }

    def __repr__(self) -> str:
        return json.dumps(self.to_dict(), indent=2)



import logging
from typing import Any, Union, Optional, Tuple

import numpy as np

from pesto.ws.core.match_apply import MatchApply, Json
from pesto.ws.features.converter.image.image import Image
from pesto.ws.features.converter.image.image_roi import ImageROI

log = logging.getLogger(__name__)


class ImageConverter(MatchApply):
    URI_STRINGS = ["file://", "http://", "https://", "gs://", "s3://"]

    def __init__(self, image_roi: Optional[ImageROI]):
        self.roi = image_roi

    def match(self, schema: Json):
        ref = schema.get("$ref")
        return ref in [
            "#/definitions/Image",
            "#/definitions/Images",
        ]

    def convert(self, data: Tuple[Any, Json]):
        payload, schema = data

        if payload is None:
            return payload

        if isinstance(payload, list):
            return [self._load_image(e) for e in payload]
        return self._load_image(payload)

    def _load_image(self, image_ref: Union[str, bytes]) -> np.ndarray:
        if isinstance(image_ref, str) and any([image_ref.startswith(uri) for uri in self.URI_STRINGS]):
            image = Image.from_uri(image_ref, roi=self.roi)
            source = "uri"
        else:
            try:
                image = Image.from_bytes(image_ref)
                source = "bytes"
            except:
                image = Image.from_base64(image_ref)
                source = "base64"

        data = image.array
        log.info("loading image : {} : shape={}".format(source, data.shape))
        return data



# from pesto.common.testing.endpoint_manager import EndpointChecker
#
#
# def test_endpoint_checker():
#     checker = EndpointChecker(
#         server_url='http://localhost:4000'
#     )
#     checker.test_with('resources/data')



import subprocess
import time
from typing import Any

import docker
import docker.errors
import requests
from pesto.common.testing import logger


class ServiceManager:
    CLIENT = docker.from_env()

    def __init__(
        self,
        docker_image: str,
        host_port: int = 4000,
        service_port: int = 8080,
        host_volume_path: str = "/tmp",
        image_volume_path: str = "/tmp",
        nvidia=False,
        attach_when_running=False,
        network: str = None
    ):
        self.docker_image = docker_image
        self.host_port = host_port
        self.service_port = service_port
        self.host_volume_path = host_volume_path
        self.image_volume_path = image_volume_path
        self.nvidia = nvidia
        self.attach_when_running = attach_when_running
        if network is None:
            self.network = "bridge"
        else:
            self.network = network

        self._container, self._existing_container = self._check_existing_container()

    def _check_existing_container(self):
        containers = ServiceManager.CLIENT.containers.list()
        for container in containers:
            ports = container.ports  # {'8080/tcp': [{'HostIp': '', 'HostPort': '4000'}]}
            tags = container.image.tags
            if self.docker_image in tags:
                for ports_dict in ports.get("{}/tcp".format(self.service_port), []):
                    if ports_dict.get("HostPort", "") == "{}".format(self.host_port):
                        logger.info("Already running container found {}".format(container.id))
                        return container, True

        return None, False

    @property
    def server_url(self):
        return "http://localhost:{}".format(self.host_port)

    @property
    def is_alive(self):
        try:
            response = requests.get("{}/api/v1/health".format(self.server_url))
            return response.status_code == 200
        except:
            return False

    def pull(self):
        try:
            self.CLIENT.images.get(self.docker_image)
        except docker.errors.ImageNotFound:
            logger.info("Pulling {}".format(self.docker_image))
            self.CLIENT.images.pull(self.docker_image)
            logger.info("Pulled image")
        except docker.errors.APIError:
            logger.info("Can't pull image {} you should have it in local other it will not work !".format(
                self.docker_image))

    def attach(self):
        if self._container is not None:
            logger.info("Attaching in subprocess to container")
            cmd = "docker attach {}".format(self._container.id)
            logger.info(cmd)
            subprocess.Popen(cmd, shell=True)
            time.sleep(5)

    def run(self):
        if self._container is None and self.is_alive:
            logger.warn("There is a container running at {} but it does not match {}.\n"
                        "We assume that you know what you are doing".format(
                self.server_url, self.docker_image))
            return None

        if self._container is None:
            self.pull()
            logger.info("Starting container with {} on port {}".format(self.docker_image, self.host_port))

            if self.network is "host":
                self._container = self.CLIENT.containers.run(
                    self.docker_image,
                    network=self.network,
                    detach=True,
                    remove=True,
                    runtime="nvidia" if self.nvidia else None,
                    volumes={self.image_volume_path: {
                        "bind": self.host_volume_path,
                        "mode": "rw",
                    }},
                )
            else:
                self._container = self.CLIENT.containers.run(
                    self.docker_image,
                    network=self.network,
                    ports={self.service_port: self.host_port},
                    detach=True,
                    remove=True,
                    runtime="nvidia" if self.nvidia else None,
                    volumes={self.image_volume_path: {
                        "bind": self.host_volume_path,
                        "mode": "rw",
                    }},
                )
            time.sleep(2)
            logger.info("Container {} started, available at {}".format(self._container.id, self.server_url))

        # Healthcheck until service responds
        num_tries = 0
        max_tries = 10

        logger.info("Trying api/v1/health for {}st time".format(num_tries + 1))
        while (num_tries < max_tries) and (not self.is_alive):
            logger.info("Server not yet alive")
            num_tries += 1
            time.sleep(2)
            logger.info("Trying api/v1/health for {}th time".format(num_tries + 1))

        if num_tries < max_tries:
            logger.info("Server alive")
        else:
            logger.error("Timeout reached")

        # Attach
        if self.attach_when_running:
            self.attach()

        return self._container.id

    def stop(self):
        if self._container is not None:
            logger.info("Stopping container {}".format(self._container.id))
            self._container.stop()

    def __enter__(self) -> "ServiceManager":
        self.run()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if not self._existing_container:
            self.stop()



import logging

import numpy as np
import rasterio

from pesto.ws.core.pesto_feature import PestoFeature
from pesto.ws.features.converter.image.bands import FullBand, Band

log = logging.getLogger(__name__)

# TODO: Classe WIP (les méthodes expérimentales sont privées)
# Potentiellement: Revoir le rationnel / faire un travail fonctionnel
# Notamment, la gestion des ROI doit-elle être dans pesto ? => Normalement le traitement devrait recevoir des
# tuiles déjà formatées et ne pas devoir gérer le tuilage (cf spam lib etc...)
# Vérifier: Pourquoi ?


class ImageROI(object):
    def __init__(self, roi: dict):
        self.target_in = roi['target_in']
        self.target_out = roi['target_out']
        self.lines = Band(roi['lines']) if 'lines' in roi else FullBand()
        self.columns = Band(roi['columns']) if 'columns' in roi else FullBand()
        self.crop_infos = (0, 0, 0, 0)

    def compute_crop_infos(self) -> PestoFeature:
        return _ComputeCropInfos(self)

    def remove_margin(self) -> PestoFeature:
        return _RemoveMargin(self)


class DummyImageROI(ImageROI):
    """
    This class is just a placeholder to initialize pipelines when no roi is provided ...
    """

    def __init__(self):
        """
        do not call super().__init__(roi) because no roi is provided
        """
        pass

    def compute_crop_infos(self):
        return None

    def remove_margin(self):
        return None


class _ComputeCropInfos(PestoFeature):
    def __init__(self, roi: ImageROI):
        self.roi = roi

    def process(self, payload: dict) -> dict:
        self.check_payload(payload)
        for _ in self.roi.target_in:
            log.info('ROI preprocess: [{}]'.format(_))
            image_path = payload[_]
            with rasterio.open(image_path) as dataset:
                shape = dataset.shape
            top, bottom = self.roi.lines.compute_crop(shape[0])
            left, right = self.roi.columns.compute_crop(shape[1])
            self.roi.crop_infos = (top, bottom, left, right)
        return payload

    def check_payload(self, payload: dict):
        ref = None
        for _ in self.roi.target_in:
            image_path = payload[_]
            with rasterio.open(image_path) as dataset:
                shape = dataset.shape
                if ref is None:
                    ref = shape
            if shape != ref:
                raise ValueError('All images should have the same shape {} ! Wrong shape is {}'.format(ref, shape))


class _RemoveMargin(PestoFeature):
    def __init__(self, roi: ImageROI):
        self.roi = roi

    def process(self, payload: dict) -> dict:
        for x in payload:
            log.info('payload: {}'.format(x))
        top, bottom, left, right = self.roi.crop_infos

        for _ in self.roi.target_out:
            log.info('ROI postprocess: [{}]'.format(_))
            image = payload[_]
            h, w = _get_shape(image)
            payload[_] = image[..., top:h - bottom, left:w - right]
        return payload


def _get_shape(image: np.ndarray):
    if len(image.shape) == 2:
        return image.shape
    return image.shape[1:3]



import os
import shlex
import subprocess
from typing import List

from jinja2 import Environment, FileSystemLoader

from pesto.cli import PROCESSING_FACTORY_PATH
from pesto.cli.core.build_config import BuildConfig
from pesto.cli.core.utils import PESTO_LOG


class DockerBuilder(object):

    @staticmethod
    def load_template():
        env = Environment(
            loader=FileSystemLoader(os.path.join(PROCESSING_FACTORY_PATH, 'pesto/cli/resources')),
            trim_blocks=True,
            variable_start_string='${',
            variable_end_string='}'
        )
        template = env.get_template('Dockerfile')
        return template

    def __init__(self, requirements: dict, build_config: BuildConfig):
        self.build_config = build_config
        self.algo_name = build_config.name
        self.base_image = requirements['dockerBaseImage']
        self.requirements = requirements['requirements']
        self.environments = requirements['environments']

    def build(self, path: str) -> None:
        dockerfile = self.dockerfile()

        path = os.path.join(path, 'Dockerfile')
        with open(path, 'w+') as file:
            file.write(dockerfile)

        docker_image_name = self.build_config.docker_image_name
        cmd = "docker build --no-cache"
        if self.build_config.network is not None:
            cmd = "{} --network='{}'".format(cmd, self.build_config.network)
        cmd = "{} -t {} {}".format(cmd, docker_image_name, self.build_config.workspace)
        subprocess.call(shlex.split(cmd))

    def dockerfile(self):
        template = self.load_template()
        return template.render(
            base_image=self.base_image,
            algo_name=self.algo_name,
            pip_extra_index=self.build_config.pip_extra_index,
            pip_proxies=self.build_config.pip_proxies,
            env_variables=self._env_variables(),
            pip_requirements=self._pip_requirements(),
            resources_requirements=self._resources()
        )

    def _env_variables(self):
        return {
            'PESTO_PROFILE': self.build_config.full_version,
            **self.environments,
            'PYTHONPATH': "".join(["$PYTHONPATH${PYTHONPATH:+:}", self._python_path])
        }

    @property
    def _python_path(self) -> str:
        return ':'.join([
            '/opt/{}'.format(self.algo_name),
            *[self.requirements[_]['to'] for _ in self._filter_requirements(['python'])]
        ])

    def _filter_requirements(self, types: List[str], include: bool = True):
        def select(req: dict):
            if include:
                return self.requirements[req].get('type', None) in types
            return self.requirements[req].get('type', None) not in types

        return filter(select, self.requirements)

    def _pip_requirements(self):
        pip_requirements = self._filter_requirements(['pip'])
        output = [os.path.basename(self.requirements[_]['from']) for _ in pip_requirements]
        PESTO_LOG.info('pip requirements : {}'.format(output))
        return output

    def _resources(self):
        resources_requirements = self._filter_requirements(['pip'], include=False)
        output = {(name, self.requirements[name]['to']) for name in resources_requirements}
        PESTO_LOG.info('resources requirements : {}'.format(output))
        return output



import os
from pesto.ws.features.converter.image.image import Image
from pesto.ws.features.converter.image.image_roi import ImageROI


def create_payload(image_path: str):
    return {
        'img1': image_path,
        'img2': image_path,
        "pesto": {
            "roi": {
                "target_in": ['img1', 'img2'],
                "target_out": ['img1', 'img2'],
                "lines": {
                    "first": 10,
                    "number": 20,
                    "margin": 20
                },
                "columns": {
                    "first": 190,
                    "number": 56,
                    "margin": 50
                }
            }
        }
    }


def test_roi():
    payload = create_payload(os.path.join(os.path.dirname(__file__), 'resources/img_256x128.jpg'))

    pesto = payload.pop('pesto')
    roi_conf = pesto.pop('roi')
    roi = ImageROI(roi_conf)

    payload = roi.compute_crop_infos().process(payload)

    # load images with cropping
    for _ in roi.target_in:
        payload[_] = Image.from_uri(payload[_], roi).array
        shape = payload[_].shape
        assert shape == (3, 50, 116)

    # remove margins
    payload = roi.remove_margin().process(payload)

    for _ in roi.target_out:
        shape = payload[_].shape
        assert shape == (3, roi.lines.number, roi.columns.number)



from typing import Any

from pesto.ws.core.pesto_feature import PestoFeature


class AlgorithmWrapper(PestoFeature):
    def __init__(self, algorithm: Any):
        self.algorithm = algorithm

    def process(self, payload: dict) -> dict:
        payload = self.algorithm.process(**payload)
        return payload



from pathlib import Path

from pesto.common.testing.test_runner import TestRunner

port = 4000
service = "{{cookiecutter.project_sname}}:{{cookiecutter.project_version}}"
# service = "{{cookiecutter.project_sname}}:{{cookiecutter.project_version}}-stateful" # test stateful profile
nvidia = "gpu" in service

test_dir = Path(__file__).parent

print("Running in {}".format(test_dir))

# Mimic Pesto Test
results = TestRunner(docker_image_name=service, nvidia=nvidia).run_all(test_dir / "resources")

ok_dict = {"NoDifference": True}


def test_describe():
    assert results.get("describe") == ok_dict


def test_processing():
    for test_key in results.keys():
        if "test_" in test_key:
            assert results.get(test_key) == ok_dict



from pathlib import Path
import json

from pesto.common.testing.endpoint_manager import EndpointManager
from pesto.common.testing.payload_generator import PayloadGenerator
from pesto.common.testing.service_manager import ServiceManager
from pesto.common.testing.service_tester import ServiceTester

port = 4000
service = "{{cookiecutter.project_sname}}:{{cookiecutter.project_version}}"
# service = "{{cookiecutter.project_sname}}:{{cookiecutter.project_version}}-stateful" # test stateful profile
nvidia = "gpu" in service

test_dir = Path(__file__).parent.parent / "pesto" "/tests"

print("Running in {}".format(test_dir))

test_files_in = test_dir / "resources" / "test_1" / "input"
test_files_out = test_dir / "resources" / "test_1" / "output"

with ServiceManager(docker_image=service, host_port=port, nvidia=nvidia) as service:
    # # Attach to logs in a subprocess
    # service.attach()

    # Wrapper around pesto endpoints to manage get & post requests
    endpoints = EndpointManager(server_url=service.server_url)

    # Tester utils to compare payloads / responses against expected values
    service_tester = ServiceTester(server_url=service.server_url)

    # check that the server is alive
    print(endpoints.is_alive)

    # get describe
    print(endpoints.describe)

    # validate describe
    with open(test_dir / "resources" / "expected_describe.json", "r") as f:
        expected = json.load(f)
        diff = service_tester.validate_describe(expected)

    print(diff)

    # generate input payload
    payload = PayloadGenerator(images_as_base64=True).generate(str(test_files_in))

    # generate expected response payload
    expected = PayloadGenerator(images_as_base64=False).generate(str(test_files_out))

    # send processing request with payload
    response = endpoints.process(payload)
    print(response)

    # validate response against expected response
    response, results = service_tester.validate_process(payload, expected)

    print(results)



import filecmp
import json
import os
from tempfile import NamedTemporaryFile
from typing import Any

import jsonschema
# TODO: fix this import (should not import anything from ws)
from pesto.ws.features.converter.image.image import Image


def mkdir(path: str) -> None:
    directory = os.path.dirname(path)
    if not os.path.exists(directory):
        os.makedirs(directory, exist_ok=True)


def load_json(path: str, filename: str = None) -> Any:
    if filename is not None:
        path = os.path.join(path, filename)
    with open(path) as file:
        json_content = json.load(file)
    return json_content


def validate_json(dictionary: dict, schema: dict) -> dict:
    jsonschema.validate(dictionary, schema)
    return dictionary


def save_json(path: str, item: Any) -> str:
    os.makedirs(os.path.dirname(path), exist_ok=True)

    with open(path, 'w') as outfile:
        json.dump(item, outfile, indent=2, sort_keys=True)
    return path


def truncate_dict_for_debug(d: dict, max_size=80):
    dd = dict()
    for k in d:
        val = d[k]
        if isinstance(val, dict):
            dd[k] = truncate_dict_for_debug(val)
        elif isinstance(val, str):
            dd[k] = val[:max_size] + ("..." if len(d[k]) > max_size else "")
        elif isinstance(val, list):
            dd[k] = [(v[:max_size] + ("..." if len(d[k]) > max_size else "") if isinstance(v, str) else v) for v in val]
        else:
            dd[k] = val

    return dd


def compare_dicts(expected: dict, actual: dict) -> dict:
    def _is_file(val):
        return isinstance(val, str) and (os.path.exists(val) or os.path.exists(val.replace("file://", "")))

    def _is_image(name: str) -> bool:
        return name.endswith(".tif") or name.endswith(".jpg") or name.endswith(".png")

    def _compare_vals(expected_v: Any, actual_v: Any) -> bool:

        if _is_file(expected_v) and _is_file(actual_v):
            expected_path = expected_v.replace("file://", "")
            response_path = actual_v.replace("file://", "")
            return filecmp.cmp(response_path, expected_path)
        elif (not _is_file(actual_v)) and (_is_file(expected_v) and _is_image(expected_v)):
            response_path = "{}{}".format(NamedTemporaryFile().name, os.path.splitext(expected_v)[1])
            response_path = Image.from_base64(actual_v).to_path(response_path)
            response_path = response_path.replace("file://", "")
            expected_path = expected_v.replace("file://", "")
            return filecmp.cmp(response_path, expected_path)
        elif isinstance(expected_v, list):
            return all([any([_compare_vals(v1, v2) for v2 in actual_v]) for v1 in expected_v])
        else:
            return expected_v == actual_v

    expected_keys = set(expected.keys())
    actual_keys = set(actual.keys())
    shared_keys = expected_keys.intersection(actual_keys)

    added = list(actual_keys - shared_keys)
    removed = list(expected_keys - shared_keys)

    modified = dict()

    for o in shared_keys:
        exp_v, actual_v = expected[o], actual[o]
        if isinstance(exp_v, dict) and isinstance(actual_v, dict):
            if not exp_v == actual_v:
                _cmp = compare_dicts(exp_v, actual_v)
                if _cmp is not None:
                    modified[o] = _cmp
        elif not _compare_vals(exp_v, actual_v):
            modified[o] = {
                "actual": actual_v,
                "expected": exp_v
            }

    comparison = dict()

    if len(added) > 0:
        comparison["KeysNotExpected"] = added
    if len(removed) > 0:
        comparison["KeysMissing"] = removed
    if len(modified.keys()) > 0:
        comparison["KeysNotEqual"] = modified
    if len(comparison.keys()) == 0:
        return None

    return comparison



import base64
import io
import os
import numpy as np
import rasterio
from PIL import Image as PILImage
from pesto.ws.features.converter.image.image import Image as PestoImage
from pesto.ws.features.converter.image.image_roi import ImageROI

# Constants

nir_file_tif = os.path.join(os.path.dirname(__file__), "resources", "TILE_RGBN.tif")

rgb_file_tif = "/tmp/TILE_RGB.tif"
rgb_file_png = "/tmp/TILE_RGB.png"
rgb_file_jpg = "/tmp/TILE_RGB.jpg"

# Functions
def scale_chw_matrix(matrix: np.ndarray, min_percentile=1, max_percentile=99):
    try:
        c, h, w = matrix.shape
    except ValueError:
        h, w = matrix.shape
        c = 1
    matrix = matrix.reshape((c, h * w)).astype(np.float64)
    # Get 2nd and 98th percentile
    mins = np.nanpercentile(matrix, min_percentile, axis=1)
    maxs = np.nanpercentile(matrix, max_percentile, axis=1)
    matrix = (matrix - mins[:, None]) / (maxs[:, None] - mins[:, None])
    matrix = matrix.reshape((c, h, w))
    matrix = matrix.clip(0., 1.)
    return matrix


def generate_groundtruths():
    # Generate RGB 8 bits groundtruth
    with rasterio.open(nir_file_tif, "r") as dataset:
        meta = dataset.meta
        img = dataset.read()
        img = img[:3, :, :]
        img = scale_chw_matrix(img)
        img = (255. * img).astype(np.uint8)

    meta.update({"count": 3, "dtype": np.uint8})

    with rasterio.open(rgb_file_tif, "w", **meta) as dataset_rgb:
        dataset_rgb.write(img)

    pil_img = np.transpose(img, (1, 2, 0))
    pil_img = PILImage.fromarray(pil_img)
    pil_img.save(rgb_file_png, format="PNG")
    pil_img.save(rgb_file_jpg, format="JPEG")


generate_groundtruths()

# Load groundtruth
with rasterio.open(rgb_file_tif, "r") as dataset_rgb:
    img_rgb_true = dataset_rgb.read()

with rasterio.open(nir_file_tif, "r") as dataset_rgbn:
    img_rgbn_true = dataset_rgbn.read()


def test_uri_rio():
    pesto_uri_rgb = PestoImage.from_uri("file://" + rgb_file_tif).to_array()
    pesto_uri_rgbn = PestoImage.from_uri("file://" + nir_file_tif).to_array()
    pesto_uri_rgb_png = PestoImage.from_uri("file://" + rgb_file_png).to_array()

    assert np.all(img_rgbn_true == pesto_uri_rgbn)
    assert np.all(img_rgb_true == pesto_uri_rgb)
    assert np.all(img_rgb_true == pesto_uri_rgb_png)

    # Ensure that the jpg driver of rasterio produces the same result as the PIL jpg decoder
    with PILImage.open(rgb_file_jpg, mode="r") as pil_jpg:
        img_rgb_jpg_true = np.asarray(pil_jpg)
        img_rgb_jpg_true = np.transpose(img_rgb_jpg_true, (2, 0, 1))

    pesto_uri_rgb_jpg = PestoImage.from_uri("file://" + rgb_file_jpg).to_array()
    assert np.all(img_rgb_jpg_true == pesto_uri_rgb_jpg)


def test_roi_rio():
    payload = {
        "target_in": nir_file_tif,
        "target_out": nir_file_tif.replace(".tif", "_out.tif"),
        "lines": {
            "first": 128,
            "number": 128,
            "margin": 32
        },
        "columns": {
            "first": 128,
            "number": 128,
            "margin": 32
        }
    }

    roi = ImageROI(roi=payload)
    pesto_uri_rgbn = PestoImage.from_uri(uri=nir_file_tif, roi=roi).to_array()
    assert np.all(img_rgbn_true[:, (128 - 32):(128 + 128 + 32), (128 - 32):(128 + 128 + 32)] == pesto_uri_rgbn)


def test_bytes_rio():
    # Test RGB
    output_profile = dict(
        driver="GTiff",
        dtype=img_rgb_true.dtype,
        count=img_rgb_true.shape[0],
        height=img_rgb_true.shape[1],
        width=img_rgb_true.shape[2],
    )

    with rasterio.MemoryFile() as memfile:
        with memfile.open(**output_profile) as dst:
            dst.write(img_rgb_true)

        bytes_buffer = memfile.read()
        pesto_bytes_rgb = PestoImage.from_bytes(bytes_buffer).to_array()

    # Test RGBN
    output_profile = dict(
        driver="GTiff",
        dtype=img_rgbn_true.dtype,
        count=img_rgbn_true.shape[0],
        height=img_rgbn_true.shape[1],
        width=img_rgbn_true.shape[2],
    )

    with rasterio.MemoryFile() as memfile:
        with memfile.open(**output_profile) as dst:
            dst.write(img_rgbn_true)

        bytes_buffer = memfile.read()
        pesto_bytes_rgbn = PestoImage.from_bytes(bytes_buffer).to_array()

    assert np.all(img_rgb_true == pesto_bytes_rgb)
    assert np.all(img_rgbn_true == pesto_bytes_rgbn)


def test_b64_rio():
    # Test RGB
    output_profile = dict(
        driver="GTiff",
        dtype=img_rgb_true.dtype,
        count=img_rgb_true.shape[0],
        height=img_rgb_true.shape[1],
        width=img_rgb_true.shape[2],
    )

    with rasterio.MemoryFile() as memfile:
        with memfile.open(**output_profile) as dst:
            dst.write(img_rgb_true)

        bytes_buffer = memfile.read()
        bytes_b64 = base64.b64encode(bytes_buffer)
        string_b64 = bytes_b64.decode('utf-8')
        pesto_b64_rgb = PestoImage.from_base64(string_b64).to_array()

    # Test RGBN
    output_profile = dict(
        driver="GTiff",
        dtype=img_rgbn_true.dtype,
        count=img_rgbn_true.shape[0],
        height=img_rgbn_true.shape[1],
        width=img_rgbn_true.shape[2],
    )

    with rasterio.MemoryFile() as memfile:
        with memfile.open(**output_profile) as dst:
            dst.write(img_rgbn_true)

        bytes_buffer = memfile.read()
        bytes_b64 = base64.b64encode(bytes_buffer)
        string_b64 = bytes_b64.decode('utf-8')
        pesto_b64_rgbn = PestoImage.from_base64(string_b64).to_array()

    assert np.all(img_rgb_true == pesto_b64_rgb)
    assert np.all(img_rgbn_true == pesto_b64_rgbn)


def test_jpg_pil():
    pil_png = PILImage.open(rgb_file_png, mode="r")
    pil_jpg = PILImage.open(rgb_file_jpg, mode="r")
    img_rgb_jpg_true = np.asarray(pil_jpg)
    img_rgb_jpg_true = np.transpose(img_rgb_jpg_true, (2, 0, 1))

    # Test equality when sending JPG buffer from PNG image
    jpg_buffer = io.BytesIO()
    pil_png.save(jpg_buffer, format='JPEG')
    jpg_bytes = jpg_buffer.getvalue()

    pesto_bytes_jpg = PestoImage.from_bytes(jpg_bytes).to_array()

    assert (np.all(pesto_bytes_jpg == img_rgb_jpg_true))

    # Test equality when sending JPG buffer from JPG image
    png_buffer = io.BytesIO()
    pil_jpg.save(png_buffer, format='PNG')
    png_bytes = png_buffer.getvalue()

    pesto_bytes_jpg = PestoImage.from_bytes(png_bytes).to_array()

    assert (np.all(pesto_bytes_jpg == img_rgb_jpg_true))

    # Test closeness of double comrpession
    buffer = io.BytesIO()
    pil_jpg.save(buffer, format='JPEG')
    pil_bytes = buffer.getvalue()

    pesto_bytes_jpg = PestoImage.from_bytes(pil_bytes).to_array()
    pil_bytes_jpg = PILImage.open(io.BytesIO(pil_bytes))
    pil_bytes_jpg = np.asarray(pil_bytes_jpg).transpose((2, 0, 1))

    assert np.all(np.isclose(pil_bytes_jpg, pesto_bytes_jpg))


# print(np.all(img_rgbn_true == pesto_bytes_rgbn))

# Test with Base64
# buffer = BytesIO()
# image.save(buffer, 'png')
# data = base64.b64encode(buffer.getvalue())
# buffer.close()
# image_id = hashlib.sha1(image_data).hexdigest()
# return data.decode("utf-8"), image_id



import logging
from abc import ABC
from typing import List

log = logging.getLogger(__name__)


class PestoFeature(ABC):
    def process(self, payload: dict) -> dict:
        raise NotImplementedError()


class PestoFeatures(PestoFeature):
    def __init__(self, features: List[PestoFeature]):
        self.features = features

    def process(self, payload: dict) -> dict:
        result = payload
        for feat in self.features:
            log.info('PESTO PROCESSING : [{}]'.format(feat.__class__.__name__))
            result = feat.process(result)
        return result



from cookiecutter.main import cookiecutter

from pesto.cli.core.utils import PESTO_LOG


def init(target, template):
    cmd = "cookiecutter {} --output-dir {}".format(template, target)
    PESTO_LOG.info(cmd)
    PESTO_LOG.info("\nPlease fill necessary information to initialize your template\n")
    res = cookiecutter(template, output_dir=target)
    PESTO_LOG.info("Service generated at {}".format(res))



from pesto.ws.features.serializer.image_serializer import ImageSerializer
import numpy as np


def test_image_serializer():
    array = (np.random.random((3, 100, 100)) * 255).astype(np.uint8)
    data = (array, 'test_image')
    ImageSerializer('/tmp/pesto/TU/', None).convert(data)



from typing import Callable, Any, Tuple

from pesto.ws.core.match_apply import MatchApply, Json


class ObjectConverter(MatchApply):
    def __init__(self, convert: Callable):
        self.delegate = convert

    def match(self, schema: Json):
        return schema.get("type") == "object"

    def convert(self, data: Tuple[Any, Json]):
        payload, schema = data
        if payload is None:
            return payload

        return {
            key: self.delegate(e, schema["properties"][key])
            for key, e in payload.items() if key in schema["properties"]
        }



import argparse
import os
from typing import Any

import pkg_resources
from pkg_resources import resource_string

from pesto.common.pesto import PESTO_WORKSPACE
from pesto.cli import build, init, list as list_builds, test
from pesto.cli.core.utils import PESTO_LOG
from pesto.version import PESTO_VERSION

ALGO_TEMPLATE_PATH = pkg_resources.get_provider('pesto.cli').__dict__['module_path'] + '/resources/template'


def display_banner() -> None:
    processing_factory_banner = resource_string('pesto.cli.resources', 'banner.txt') \
        .decode('utf-8') \
        .replace(' ---------------', ': {:10s}----'.format(PESTO_VERSION))

    PESTO_LOG.info(processing_factory_banner)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest='subcommand', title='subcommands', help='Valid subcommands')
    subparsers.required = True

    # init
    parser_init = subparsers.add_parser('init')
    parser_init.add_argument('-t', '--template', help='path to the algorithm template', default=ALGO_TEMPLATE_PATH)
    parser_init.add_argument('target', help='path to the new algorithm folder')

    # build
    parser_build = subparsers.add_parser('build')
    parser_build.add_argument('build_config', help='path to build.json')
    parser_build.add_argument('-p', '--profile', nargs='+', help='Select specific files to update',
                              default=None)
    parser_build.add_argument('--proxy', help='Define a proxy url to use during docker construction',
                              default=None)
    parser_build.add_argument('-n', '--network', help='Define a specific network for docker construction',
                              default="host")

    # test
    parser_test = subparsers.add_parser('test')
    parser_test.add_argument('build_config', help='path to build.json')
    parser_test.add_argument('-p', '--profile', nargs='+', help='Select specific files to update',
                             default=None)
    parser_test.add_argument('--nvidia',  action='store_true', default=False, help='Run docker with nvidia-runtime')
    parser_test.add_argument('-n', '--network', help='Define a specific network to run docker',
                              default=None)

    # # list builds
    # parser_list = subparsers.add_parser('list')

    return parser.parse_args()


def main() -> None:
    display_banner()
    args = parse_args()
    if args.subcommand == 'init':
        init.init(args.target, args.template)
    elif args.subcommand == 'build':
        build.build(search_build_config(args), args.profile, args.proxy, args.network)
    elif args.subcommand == 'test':
        test.test(search_build_config(args), args.profile, nvidia=args.nvidia, network=args.network)
    elif args.subcommand == 'list':
        list_builds.list_builds(PESTO_WORKSPACE)


def search_build_config(args: Any) -> str:
    build_config_path = str(args.build_config)

    PESTO_LOG.debug('search build config ...')

    if ':' in build_config_path:
        PESTO_LOG.info('search {} in PESTO build repository : {}'.format(build_config_path, PESTO_WORKSPACE))
        name, version = build_config_path.split(':')
        build_config_path = os.path.join(PESTO_WORKSPACE, name, version, name)

    if not build_config_path.endswith('.json'):
        PESTO_LOG.warning('build parameter {} is not a json ...'.format(build_config_path))
        build_config_path = os.path.join(build_config_path, 'pesto', 'build', 'build.json')
        PESTO_LOG.warning('search for build.json in {}'.format(build_config_path))
        if not os.path.exists(build_config_path):
            raise ValueError('build.json not found at {}'.format(build_config_path))

    return build_config_path


if __name__ == "__main__":
    main()



import os
import numpy as np

from pesto.ws.features.converter.image.image import Image
from pesto.ws.features.converter.image.image_roi import ImageROI


def get_image():
    return (np.random.random((3, 20, 20)) * 255.).astype(np.uint8)


def test_image_bytes_conversion():
    data = get_image()

    img = Image.from_array(data)
    img_bytes = img.to_bytes()
    img_back = Image.from_bytes(img_bytes)
    data_back = img_back.array
    data = np.moveaxis(data, 0, -1)
    data_back = np.moveaxis(data_back, 0, -1)

    assert np.array_equal(data.shape, data_back.shape)
    # assert np.array_equal(data, data_back)


def test_image_roi():
    data = os.path.abspath(os.path.join(os.path.dirname(__file__), 'resources/img_256x128.jpg'))
    data = "file://" + data

    roi = ImageROI({
        "target_in": data,
        "target_out": data,
        'lines': {
            'first': 0,
            'number': 100,
            'margin': 10
        },
        'columns': {
            'first': 0,
            'number': 200,
            'margin': 15
        }
    })
    img_full = Image.from_uri(data).array
    img = Image.from_uri(data, roi).array
    assert (img.shape[0] == img_full.shape[0])



#!/usr/bin/python3

import argparse
from pathlib import Path
import json

from pesto.common.testing.service_manager import ServiceManager

port = 4000
service = "{{cookiecutter.project_sname}}:{{cookiecutter.project_version}}"

parser = argparse.ArgumentParser(description="""Start the PESTO webservice.
Example: """ + __file__ + """ [-p gpu]
""", formatter_class=argparse.RawTextHelpFormatter)
    
parser.add_argument('-p', '--profile', nargs='+', default="", help='List of profiles to apply. Use stateless to enable the asynchronous service')
args = parser.parse_args()

if len(args.profile):
    service = service + "-" + "-".join(args.profile)

nvidia = "gpu" in service
print("Launching service " + service)

with ServiceManager(docker_image=service, host_port=port, nvidia=nvidia) as service:
    input("Press Enter to quit ...")




import json
import shutil
from pathlib import Path
import time
from pesto.common.pesto import PESTO_WORKSPACE
from pesto.common.testing import logger
from pesto.common.testing.payload_generator import PayloadGenerator
from pesto.common.testing.service_manager import ServiceManager
from pesto.common.testing.service_tester import ServiceTester

TMP_PATH = Path("/tmp/pesto/")


class TestRunner:
    def __init__(self, docker_image_name: str, network: str = None, nvidia=False):
        self.docker_image_name = docker_image_name
        self.nvidia = nvidia
        self.network = network

        image, tag = self.docker_image_name.split(":")

        self._tmp_test_resources = TMP_PATH / "tests" / image / tag
        self._test_output = Path(PESTO_WORKSPACE) / "tests" / image / tag

    def prepare_resources(self, test_resources_path: Path):

        tests = []
        describe = {}

        shutil.rmtree(self._tmp_test_resources, ignore_errors=True)
        self._tmp_test_resources.mkdir(exist_ok=True, parents=True)

        for path in sorted(test_resources_path.iterdir()):

            if path.is_dir() and "test_" in path.stem:
                test_dir = path.stem
                shutil.rmtree(self._tmp_test_resources / test_dir, ignore_errors=True)
                shutil.copytree(path, self._tmp_test_resources / test_dir)

                tests.append(self._tmp_test_resources / test_dir)

            elif path.is_file() and path.match("*describe*.json"):
                with open(path, 'r') as f:
                    describe = json.load(f)

        return tests, describe

    def run_all(self, test_resources_path: Path):
        tests, expected_describe = self.prepare_resources(test_resources_path)

        if self.nvidia:
            logger.info("Running with nvidia runtime")

        with ServiceManager(
                docker_image=self.docker_image_name,
                nvidia=self.nvidia,
                attach_when_running=True,
        ) as service:

            # force restarting service
            if service._existing_container:
                logger.debug("Force restarting service to ensure resources properly mounted")
                service.stop()
                time.sleep(10)
                service.run()
                time.sleep(5)

            service_tester = ServiceTester(server_url=service.server_url)

            all_results = dict()

            # Save describes for debug
            with open(self._tmp_test_resources / "expected_describe.json", "w") as f:
                json.dump(expected_describe, f, indent=2)

            with open(self._tmp_test_resources / "actual_describe.json", "w") as f:
                actual_describe = service_tester.endpoint_manager.describe
                json.dump(actual_describe, f, indent=2)

            # Validate describe
            describe_differences = service_tester.validate_describe(expected_describe)

            all_results['describe'] = describe_differences

            for test_dir in tests:
                payload = PayloadGenerator(images_as_base64=False).generate(str(test_dir / "input"))
                expected = PayloadGenerator(images_as_base64=False).generate(str(test_dir / "output"))

                response, validation_results = service_tester.validate_process(payload, expected)

                all_results[test_dir.stem] = validation_results

                # Serialize response for debug
                response_path = test_dir / "response"
                with open(test_dir / "response.json", "w") as f:
                    json.dump(response, f, indent=2)

                self._serialize_response(response_path, response)

        with open(self._tmp_test_resources / "results.json", "w") as f:
            json.dump(all_results, f, indent=2)

        logger.info("--- Tests Results ---")
        logger.info(json.dumps(all_results, indent=2))
        logger.info("--- Copying tests outputs to {}".format(self._tmp_test_resources))
        shutil.rmtree(self._test_output, ignore_errors=True)
        shutil.copytree(self._tmp_test_resources, self._test_output)

        return all_results

    @staticmethod
    def _serialize_response(response_path: Path, response: dict):
        def _is_uri(val):
            if isinstance(val, str) and (val.startswith("file://") or val.startswith("/")) and Path(val).exists():
                return True
            else:
                return False

        def _write_val(val, path):
            if not _is_uri(val):
                with open(path, 'w') as f:
                    if isinstance(val, dict):
                        f.write(json.dumps(val, indent=2))
                    else:
                        f.write(str(val))
            else:
                shutil.copy(val, "{}{}".format(path, Path(val).suffix))

        response_path.mkdir(exist_ok=True)
        for key in response:
            if isinstance(response[key], list):
                (response_path / key).mkdir(exist_ok=True)
                [_write_val(_val, response_path / str(key) / str(i)) for i, _val in enumerate(response[key])]
            else:
                _write_val(response[key], response_path / str(key))



from abc import ABC
from typing import Dict, Any, List

Json = Dict[str, Any]


class MatchApply(ABC):
    def match(self, schema: Json):
        raise NotImplementedError()

    def convert(self, data: Any):
        raise NotImplementedError()


class MatchPipeline:
    def __init__(self, pipeline: List[MatchApply], default: MatchApply = None):
        self.pipeline = pipeline
        self.default = default

    def convert(self, value: Any, schema: Dict):
        for _ in self.pipeline:
            if _.match(schema):
                return _.convert(value)

        if self.default is None:
            raise ValueError('Unsupported data type {} : value={}'.format(schema, str(value)))

        return self.default.convert(value)



import json

from pesto.common.testing.payload_generator import PayloadGenerator


def test_file_loader() -> None:
    loader = PayloadGenerator('file://')
    payload = loader.generate('resources/data/input')
    with open('resources/data_expected.json') as _:
        expected = json.load(_)

    assert payload.keys() == expected.keys()
    for k in payload:
        assert payload[k] == expected[k]

    assert json.dumps(payload, sort_keys=True, indent=4) == json.dumps(expected, sort_keys=True, indent=4)



import os
from typing import Dict

from pesto.cli import PROCESSING_FACTORY_PATH
from pesto.cli.core.build_config import BuildConfig
from pesto.cli.core.pesto_files import PestoFiles
from pesto.cli.core.profile_loader import ProfileLoader
from pesto.cli.core.utils import PESTO_LOG
from pesto.common.utils import load_json, validate_json
from pesto.version import PESTO_VERSION

SCHEMA_PATH = os.path.join(PROCESSING_FACTORY_PATH, 'pesto/cli/resources/schema')


class ConfigLoader(object):

    @staticmethod
    def load(build_config: BuildConfig):
        root_path = os.path.join(build_config.algorithm_path, 'pesto')

        loader = ProfileLoader(root_path, build_config.profiles)

        configs = {
            _: loader.load(_.value)
            for _ in PestoFiles.required()
        }

        configs.update({
            PestoFiles.service: ConfigLoader._describe(configs),
            PestoFiles.version: ConfigLoader._version(configs),
            PestoFiles.build: build_config.to_dict()
        })
        return configs

    @staticmethod
    def _version(configs: Dict) -> dict:
        PESTO_LOG.info('********** generate version.json **********')
        return {
            'version': configs[PestoFiles.description]['version'],
            **configs[PestoFiles.version],
            '__packaging': 'packaged with PESTO v{}'.format(PESTO_VERSION)
        }

    @staticmethod
    def _describe(configs: Dict) -> dict:
        PESTO_LOG.info('********** generate service.json **********')
        definitions = load_json(SCHEMA_PATH, 'definitions.json')

        description = {
            **configs[PestoFiles.description],
            'config': configs[PestoFiles.config_schema],
            'input': {
                **configs[PestoFiles.input_schema],
                'definitions': definitions,
            },
            'output': {
                **configs[PestoFiles.output_schema],
                'definitions': definitions,
            },
        }

        service_schema = load_json(SCHEMA_PATH, 'service_schema.json')
        validate_json(description, service_schema)

        return description



import logging
import os
from typing import Any

from pesto.common.utils import load_json

log = logging.getLogger(__name__)


class DescribeService(object):
    SERVICE_JSON_PATH = '/etc/pesto/service.json'
    VERSION_CONTENT_PATH = '/etc/pesto/version.json'

    def __init__(self, url_root: str):
        self.description = load_json(DescribeService.SERVICE_JSON_PATH)
        self.url_root = url_root

    @staticmethod
    def compute_version() -> Any:
        return load_json(DescribeService.VERSION_CONTENT_PATH)

    def compute_describe(self) -> dict:
        describe = {
            _: self.description[_] for _ in {
                'title',
                'name',
                'version',
                'family',
                'description',
                'organization',
                'email',
                'resources',
                'asynchronous'
            }
        }

        if 'batched' in self.description:
            describe['batched'] = self.description['batched']

        describe.update({
            _: self.description[_] for _ in self.description.keys() & {
                'keywords',
                'template',
                'config'
            }
        })

        describe.update({
            'input': self._build_io_schema('input'),
            'output': self._build_io_schema('output'),
            '_links': {
                'self': {
                    'relation': 'Access to describe resource',
                    'href': self._compute_endpoint('describe'),
                    'type': 'application/json',
                    'method': 'GET'
                },
                'execution': {
                    'relation': 'Processing resource',
                    'href': self._compute_endpoint('process'),
                    'type': self._compute_execution_type(),
                    'method': 'POST'
                },
                'config': {
                    'relation': 'Processing configuration',
                    'href': self._compute_endpoint('config'),
                    'type': 'application/json',
                    'method': 'GET'
                },
                'version': {
                    'relation': 'Processing version',
                    'href': self._compute_endpoint('version'),
                    'type': 'application/json',
                    'method': 'GET'
                },
                'health': {
                    'relation': 'Processing health',
                    'href': self._compute_endpoint('health'),
                    'type': 'text/plain',
                    'method': 'GET'
                }

                # metrics
                ## TODO TO BE IMPLEMENTED

                # icon
                ## TODO Unimplemented

                # license
                ## TODO TO BE IMPLEMENTED

                # documentation
                ## TODO TO BE IMPLEMENTED
                ## Maybe return the swagger UI
            }
        })

        return describe

    def _compute_endpoint(self, name: str) -> str:
        return os.path.join(self.url_root, 'api', 'v1', name)

    def _compute_execution_type(self) -> str:
        if isinstance(self.description['output'].get('content'), str):
            return self.description['output'].get('content')
        return 'Complex type, see output in describe content for more information'

    def _build_io_schema(self, field: str) -> dict:
        properties = self.description[field]

        required = properties.pop('required', [])
        definition = properties.pop('definition', {})
        schema = properties.pop('$schema', 'http://json-schema.org/draft-06/schema#')
        title = properties.pop('title', '')
        type = properties.pop('type', "object")
        content = properties.pop('content', None)
        definitions = properties.pop('definitions', {})
        description = properties.pop('description', "Expected format")
        if 'properties' in properties:
            properties = properties['properties']

        result = {
            "$schema": schema,
            "title": title,
            "type": type,
            "description": description,
            "definition": definition,
            "definitions": definitions,
            "properties": properties
        }
        if required:
            result['required'] = required
        if content:
            result['content'] = content

        return result



class ProcessingError(Exception):

    def __init__(self, message: str) -> None:
        self.message = message

    def __repr__(self) -> str:
        return 'ProcessingError: {}'.format(self.message)



import logging
import os

from pesto.ws.core.match_apply import MatchPipeline
from pesto.ws.core.pesto_feature import PestoFeature
from pesto.ws.features.serializer.image_serializer import ImageSerializer
from pesto.ws.features.serializer.json_serializer import JsonSerializer
from pesto.ws.features.serializer.number_serializer import NumberSerializer
from pesto.ws.features.serializer.string_serializer import StringSerializer
from pesto.ws.service.job_list import JobListService

log = logging.getLogger(__name__)


class ResponseSerializer(PestoFeature):

    def __init__(self, schema: dict, job_id: str):
        self.schema = schema

        job_path = os.path.join(JobListService.PESTO_WORKSPACE, job_id)
        self.output_serialize = MatchPipeline([
            ImageSerializer(job_path, schema=schema),
            JsonSerializer(job_path),
            NumberSerializer(job_path),
            StringSerializer(job_path)
        ], default=JsonSerializer(job_path))

    def process(self, payload: dict) -> dict:
        log.info('saving response to disk ...')

        for key, schema in self.schema['properties'].items():
            ref = schema.get('$ref')
            key_type = schema.get('type')

            log.info('saving {}: $ref={} type={}'.format(key, ref, key_type))
            if payload.get(key) is not None:
                self.output_serialize.convert((payload.get(key), key), schema)

        log.info('response saved !')
        return payload



import asyncio
import time
from concurrent.futures import ThreadPoolExecutor

from sanic import Sanic
from sanic import response
from sanic.exceptions import ServerError

model = None


def processing():
    global model
    if not model:
        model = 'xxxx'
        print('loading model')

    print("start processing")
    for i in range(10):
        time.sleep(1)
        print(i)
    return {"processing": "result"}


app = Sanic()

processing_semaphore = None


@app.listener('before_server_start')
async def init(sanic, loop):
    global processing_semaphore
    processing_semaphore = asyncio.Semaphore(1, loop=loop)


async def async_exec(callback):
    with ThreadPoolExecutor(max_workers=1) as executor:
        return await asyncio.get_event_loop().run_in_executor(executor, callback)


@app.route("/process")
async def process(request):
    if processing_semaphore.locked():
        raise ServerError('a processing is already running', status_code=423)

    await processing_semaphore.acquire()
    try:
        future = async_exec(processing)
        result = await asyncio.wait_for(future, timeout=None)
        return response.json(result)
    finally:
        processing_semaphore.release()


@app.route("/health")
async def health(request):
    return response.json({"health": "ok"})


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=4000)



from setuptools import setup, find_packages

with open('README.md', "r") as readme_file:
    readme = str(readme_file.read())

with open('requirements.txt', 'r') as requirements_file:
    install_requires = [line for line in requirements_file]

setup(
    name="{{cookiecutter.project_sname}}",
    version="{{cookiecutter.project_version}}",
    description="{{cookiecutter.project_short_description}}",
    long_description=readme + '\n\n',
    author="{{cookiecutter.maintainer_fullname}}",
    author_email='{{cookiecutter.maintainer_email}}',
    url="",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    install_requires=install_requires,
    zip_safe=False,
    keywords=["pesto-project"],
)



from typing import Optional

import jsonschema
from pesto.common.testing import logger
from pesto.common.testing.endpoint_manager import EndpointManager
from pesto.common.utils import compare_dicts


class ServiceTester:
    def __init__(self, server_url):
        self.server_url = server_url
        self.endpoint_manager = EndpointManager(self.server_url)

    def validate_health(self):
        return self.endpoint_manager.is_alive

    def validate_describe(self, expected_describe: dict):
        describe = self.endpoint_manager.describe
        expected = expected_describe

        comparison_results = compare_dicts(expected, describe)

        comparison_results = comparison_results or {"NoDifference": True}

        return comparison_results

    def validate_process(self, payload: dict, expected_response: Optional[dict]):

        validation_result = dict()

        logger.info("Validating payload against input schema")
        try:
            jsonschema.validate(payload, self.endpoint_manager.input_schema)
        except jsonschema.ValidationError as e:
            print(e)
            validation_result["input_schema"] = False
            return "NO RESPONSE", validation_result
        except jsonschema.SchemaError as e:
            print(e)
            validation_result["input_schema"] = False
            return "NO RESPONSE", validation_result

        logger.info("Validation OK")
        validation_result["input_schema"] = True

        logger.info("Computing response")
        response = self.endpoint_manager.process(payload)

        logger.info("Validating response against output schema")
        try:
            jsonschema.validate(payload, self.endpoint_manager.input_schema)
        except jsonschema.ValidationError as e:
            logger.error(e)
            validation_result["output_schema"] = False
        except jsonschema.SchemaError as e:
            logger.error(e)
            validation_result["output_schema"] = False

        logger.info("Validation OK")
        validation_result["output_schema"] = True
        expected_response = expected_response or dict()

        validation_result = compare_dicts(expected_response, response)

        validation_result = validation_result or {"NoDifference": True}

        return response, validation_result



import json
import os
import shlex
import shutil
import subprocess
import tarfile
import glob
from pathlib import Path
from typing import List

import wget

from pesto.cli import PROCESSING_FACTORY_PATH
from pesto.cli.core.build_config import BuildConfig
from pesto.cli.core.config_loader import ConfigLoader
from pesto.cli.core.docker_builder import DockerBuilder
from pesto.cli.core.pesto_files import PestoFiles
from pesto.cli.core.utils import PESTO_LOG
from pesto.common.utils import load_json, validate_json, mkdir


def copy_requirement(from_uri: str, target_path: str) -> None:
    mkdir(target_path)
    if from_uri.startswith("gs://"):
        PESTO_LOG.info('copy to : {}'.format(target_path))
        subprocess.call(shlex.split('gsutil cp {0} {1}'.format(from_uri.rstrip(os.path.sep), target_path)))
    elif from_uri.startswith("file://") and os.path.exists(from_uri.replace("file://", "")):
        shutil.copyfile(from_uri.replace("file://", ""), target_path)
    else:
        wget.download(url=from_uri, out=target_path)


class Builder:
    SCHEMA_PATH = os.path.join(PROCESSING_FACTORY_PATH, 'pesto/cli/resources/schema')

    def __init__(self, build_config: BuildConfig):
        PESTO_LOG.info('***** init packaging *****')
        self.build_config = build_config
        self.configs = ConfigLoader.load(build_config)

        PESTO_LOG.info('********** build parameters **********')
        PESTO_LOG.info('{}'.format(json.dumps(build_config.__dict__, indent=4)))
        PESTO_LOG.info('processing factory path : {}'.format(PROCESSING_FACTORY_PATH))

    def conf_validation(self) -> None:
        PESTO_LOG.info('********** validate requirements.json **********')
        requirements_schema = load_json(Builder.SCHEMA_PATH, 'requirements_schema.json')
        validate_json(self.configs[PestoFiles.requirements], requirements_schema)

        PESTO_LOG.info('********** validate config.json **********')
        validate_json(self.configs[PestoFiles.config], self.configs[PestoFiles.config_schema])

    def copy_factory_files(self) -> None:
        PESTO_LOG.info('********** copy factory files **********')
        if os.path.exists(self.workspace):
            shutil.rmtree(self.workspace)
        os.makedirs(self.workspace, exist_ok=True)

        PESTO_LOG.info('workspace created : {}'.format(self.workspace))

        # copy pesto required resources (api_geo_process_v1.0.yaml)
        PESTO_LOG.debug('COPY pesto resources to workspace from: {}'.format(PROCESSING_FACTORY_PATH))
        # shutil.copytree(os.path.join(PROCESSING_FACTORY_PATH, 'pesto/cli'), os.path.join(self.workspace, 'pesto/cli'))
        os.makedirs(os.path.join(self.workspace, "pesto"), exist_ok=True)
        shutil.copyfile(os.path.join(PROCESSING_FACTORY_PATH, 'pesto/cli/resources/doc/api_geo_process_v1.0.yaml'),
                        os.path.join(self.workspace, 'pesto/api_geo_process_v1.0.yaml'))

        # copy algorithm
        target_path = os.path.join(self.workspace, self.build_config.name)
        PESTO_LOG.debug('COPY algorithm to workspace from: {} to: {}'.format(
            self.build_config.algorithm_path,
            target_path))
        shutil.copytree(self.build_config.algorithm_path, target_path)

        # copy/update config files
        shutil.rmtree(os.path.join(target_path, 'pesto', 'api'))
        shutil.rmtree(os.path.join(target_path, 'pesto', 'build'))
        os.makedirs(os.path.join(target_path, 'pesto', 'api'))
        os.makedirs(os.path.join(target_path, 'pesto', 'build'))

        for item in self.configs:
            with open(os.path.join(target_path, 'pesto', item.value), 'w') as _:
                json.dump(self.configs[item], _, indent=4, sort_keys=True)

    def copy_requirements(self) -> None:
        PESTO_LOG.info('********** copy requirements **********')
        requirements = self.configs[PestoFiles.requirements]['requirements']
        for name in requirements.keys():
            from_uri = requirements[name]['from']
            temporary_path = os.path.join(self.workspace, 'requirements', os.path.basename(from_uri))
            target_path = os.path.join(self.workspace, name)

            PESTO_LOG.info('COPY from {} to {}'.format(from_uri, temporary_path))
            copy_requirement(from_uri, temporary_path)

            if from_uri.endswith('tar.gz'):
                PESTO_LOG.info('EXTRACT from {} to {}'.format(temporary_path, target_path))
                with tarfile.open(temporary_path, 'r:gz') as file:
                    file.extractall(path=target_path)

    def copy_pesto_whl(self) -> None:
        PESTO_LOG.info('********** copy local pesto wheel if any **********')
        source_dir = os.path.join(Path.home(), ".pesto/dist")
        dest_dir = os.path.join(self.workspace, 'dist/')
        os.makedirs(dest_dir, exist_ok=True)
        for filename in glob.glob(os.path.join(source_dir, '*.*')):
            PESTO_LOG.info('********** copy {} in {} **********'.format(filename, dest_dir))
            shutil.copy(filename, dest_dir)

    def build_docker_image(self) -> None:
        PESTO_LOG.info('********** build docker image **********')
        DockerBuilder(self.configs[PestoFiles.requirements], self.build_config).build(self.workspace)

    @property
    def workspace(self):
        return self.build_config.workspace


def build(build_config_path: str, profiles: List[str], proxy: str = None, network: str = "host") -> None:
    config = BuildConfig.from_path(path=build_config_path, profiles=profiles, proxy=proxy, network=network)

    builder = Builder(config)
    builder.conf_validation()
    builder.copy_factory_files()
    builder.copy_requirements()
    builder.copy_pesto_whl()
    builder.build_docker_image()



import logging
import os
from typing import Tuple, Any, Dict

import numpy as np
from pesto.ws.core.match_apply import MatchApply, Json
from pesto.ws.features.converter.image.image import Image

log = logging.getLogger(__name__)


class ImageSerializer(MatchApply):
    def __init__(self, job_path: str, schema: Dict):
        self.job_path = job_path
        try:
            output_content = schema["content"]["content"]
            self._default_driver = output_content.replace("image/", "")
        except:
            self._default_driver = "png"

    def match(self, schema: Json):
        ref = schema.get("$ref")
        return ref in [
            "#/definitions/Image",
            "#/definitions/Images",
        ]

    def convert(self, data: Tuple[Any, str]):
        payload, key = data
        if isinstance(payload, list):
            for index, array in enumerate(payload):
                self._save_image(array, key + str(index))
            return
        array = payload
        self._save_image(array, key)

    def _save_image(self, array: np.ndarray, key: str) -> None:
        image = Image(array=array)

        if image.bands() > 4:
            driver = "tif"
        else:
            driver = self._default_driver

        filename = os.path.join(self.job_path, "{}.{}".format(key, driver))
        log.info("saving image : " + filename)
        image.to_path(filename)



from pathlib import Path

from pesto.cli.core.build_config import BuildConfig
from pesto.cli.core.utils import PESTO_LOG
from pesto.common.testing.test_runner import TestRunner


def test(build_config_path, profiles, nvidia=False, network=None):
    build_config = BuildConfig.from_path(path=build_config_path, profiles=profiles, network=network)
    PESTO_LOG.info('build configuration : {}'.format(build_config))

    pesto_path = Path(build_config.algorithm_path) / 'pesto' / 'tests' / 'resources'

    TestRunner(docker_image_name=build_config.docker_image_name, network=network, nvidia=nvidia).run_all(pesto_path)



import logging
import os
import shutil

from pesto.ws.service.job_list import JobListService

log = logging.getLogger(__name__)


class JobDeleteService:
    def __init__(self, url_root: str, job_id: str):
        self.job_id = job_id
        self.job_path = os.path.join(JobListService.PESTO_WORKSPACE, self.job_id)
        self.url_root = url_root

    def delete(self) -> None:
        log.info('delete: job_id = {}'.format(self.job_id))
        shutil.rmtree(self.job_path)

    def delete_partial(self, result_id: str) -> None:
        for ext in [
            'json',
            'tif', 'jpeg', 'png',
            'float', 'int', 'string'
        ]:
            _remove_silent(os.path.join(self.job_path, result_id + '.' + ext))


def _remove_silent(path: str) -> None:
    try:
        os.remove(path)
    except:
        pass



import logging
import os
import shutil
from typing import Any

from pesto.ws.features.converter.image.image import Image
from pesto.ws.core.pesto_feature import PestoFeature
from pesto.ws.service.job_list import JobListService
from pesto.ws.service.job_result import JobResultService, ResultType

log = logging.getLogger(__name__)


class StatelessResponse(PestoFeature):
    def __init__(self, url_root: str, job_id: str, schema: dict):
        self.result_service = JobResultService(url_root, job_id)
        self.url_root = url_root
        self.job_id = job_id
        self.schema = schema
        self.output_properties = self.schema['properties']

    def process(self, payload: dict) -> Any:
        try:
            if self._unique_response():
                log.info('response mode : stateless : unique response')
                result_id = list(self.schema.keys())[0]
                path, data_type = self.result_service.get_partial_result(result_id)
                if data_type == ResultType.image:
                    return path, ResultType.image

            log.info('response mode : stateless')
            for key in payload.keys():
                key_type = self.output_properties[key].get('$ref')
                if key_type == '#/definitions/Image':
                    payload[key] = Image.from_array(payload[key]).to_base64()
                elif key_type == '#/definitions/Images':
                    for index, img in enumerate(payload[key]):
                        payload[key][index] = Image.from_array(payload[key][index]).to_base64()
            return payload, ResultType.json
        finally:
            job_path = os.path.join(JobListService.PESTO_WORKSPACE, self.job_id)
            shutil.rmtree(job_path, ignore_errors=True)

    def _unique_response(self) -> bool:
        log.info('output_properties: {}'.format(str(self.output_properties)))

        try:
            unique_response = len(set(self.output_properties)) == 1
            key_type = self.output_properties[0].get('$ref')
            is_dict = key_type in ['#/definitions/Metadata',
                                   '#/definitions/Metadatas',
                                   '#/definitions/Polygon',
                                   '#/definitions/Polygons'
                                   ]

            return unique_response and is_dict
        except:
            return False



import logging
import os
from enum import Enum
from typing import List, Tuple, Optional
import re
from pesto.common.utils import load_json
from pesto.ws.service.job_list import JobListService

log = logging.getLogger(__name__)


class ResultType(Enum):
    image = ['tif', 'png', 'jpg']
    file = ['string', 'float', 'int']
    json = ['json']

    @property
    def extensions(self):
        return self.value

    def transform(self, path: str) -> str:
        if self == ResultType.json:
            return load_json(path)
        if self == ResultType.file:
            with open(path, 'r') as f:
                return f.read()
        if self == ResultType.image:
            return path


class JobResultService:

    def __init__(self, url_root: str, job_id: str):
        self.job_id = job_id
        self.job_path = os.path.join(JobListService.PESTO_WORKSPACE, self.job_id)
        self.url_root = url_root

    def get_results(self) -> dict:
        result = {}
        for file in sorted(os.listdir(self.job_path)):
            name, ext = os.path.splitext(file)
            if re.search(r'\d+$', name) is not None:
                idx = re.search(r'\d+$', name).group()
                if name.endswith(idx):
                    new_name = name[:-len(idx)]
                    if result.get(new_name):
                        result[new_name].append('{}/api/v1/jobs/{}/results/{}'.format(self.url_root, self.job_id, name))
                    else:
                        result[new_name] = ['{}/api/v1/jobs/{}/results/{}'.format(self.url_root, self.job_id, name)]
            else:
                if name != '__response':
                    result[name] = '{}/api/v1/jobs/{}/results/{}'.format(self.url_root, self.job_id, name)

        return result

    def get_partial_result(self, result_id: str) -> Tuple[str, ResultType]:
        for data_type in ResultType:
            path = self._get_partial_result_path(data_type.extensions, result_id)
            if path is not None:
                output = data_type.transform(path)
                log.info('get job result {} : type={} output={}'.format(result_id, data_type, output))
                return output, data_type

        raise ValueError('No partial result found for job_id={} result_id={}'.format(self.job_id, result_id))

    def _get_partial_result_path(self, extensions: List[str], result_id: str) -> Optional[str]:
        for extension in extensions:
            path = os.path.join(self.job_path, result_id + '.' + extension)
            if os.path.exists(path):
                return path

        return None



import logging
import os

log = logging.getLogger(__name__)


class JobListService(object):
    PESTO_WORKSPACE = '/tmp/.pesto/jobs'

    def __init__(self) -> None:
        self.PESTO_WORKSPACE = JobListService.PESTO_WORKSPACE

    def job_list(self, url_root: str) -> dict:
        log.info('job_list : url_root = {}'.format(url_root))

        result = {}

        if os.path.exists(self.PESTO_WORKSPACE):
            log.info('workspace found : {}'.format(self.PESTO_WORKSPACE))
            for job_id in os.listdir(self.PESTO_WORKSPACE):
                result[job_id] = {'link': '{}/api/v1/jobs/{}/status'.format(url_root, job_id)}
        return result



import json
import os
from typing import Any, Dict, Callable, List, Union

from pesto.ws.features.converter.image.image import Image


class PathLoader:
    def match(self, file_path: str) -> bool:
        raise NotImplementedError()

    def load(self, file_path: str) -> Any:
        raise NotImplementedError()


# DEFAULT (URI based) loader ################################################


class UriLoader(PathLoader):
    def __init__(self, file_prefix: str):
        self.file_prefix = file_prefix

    def match(self, file_path: str) -> bool:
        return True

    def load(self, file_path: str) -> Any:
        return "{}{}".format(self.file_prefix, file_path)


# Recursive (FOLDER) loader #################################################


class FolderLoader(PathLoader):
    def __init__(self, global_loader: Callable[[str], Any]) -> None:
        self.loader = global_loader

    def match(self, file_path: str) -> bool:
        return os.path.isdir(file_path)

    def load(self, file_path: str) -> Union[Dict, List]:
        is_array = all([os.path.splitext(x)[0].isdigit() for x in os.listdir(file_path)])

        if is_array:
            return [self.loader(os.path.join(file_path, x)) for x in sorted(os.listdir(file_path))]
        else:
            return {
                os.path.splitext(x)[0]: self.loader(os.path.join(file_path, x))
                for x in sorted(os.listdir(file_path))
            }


# Specific FILE loaders ####################################################


class FileLoader(PathLoader):
    def __init__(self, *args: str):
        self.extension_pattern: [str] = [arg.lower() for arg in args]

    @staticmethod
    def _match(file_path, extension_pattern):
        return file_path.lower().endswith(extension_pattern)

    def match(self, file_path: str) -> bool:
        return any([self._match(file_path, ext) for ext in self.extension_pattern])

    def load(self, file_path: str) -> Any:
        raise NotImplementedError()


class Base64ImageLoader(FileLoader):
    def load(self, file_path: str) -> str:
        image = Image.from_uri(file_path)
        return image.to_base64()


class StringLoader(FileLoader):
    def load(self, file_path: str) -> str:
        with open(file_path) as _:
            return str(_.read())


class FloatLoader(FileLoader):
    def load(self, file_path: str) -> float:
        with open(file_path) as _:
            return float(_.read())


class IntLoader(FileLoader):
    def load(self, file_path: str) -> int:
        with open(file_path) as _:
            return int(_.read())


class JsonLoader(FileLoader):
    def load(self, file_path: str) -> Any:
        with open(file_path) as _:
            return json.load(_)


# Generic loader based on delegation to a sequence of loaders ##############


class GenericLoader(PathLoader):
    def __init__(self) -> None:
        self.loaders: List[PathLoader] = []

    def set_loaders(self, loaders: List[PathLoader]) -> None:
        self.loaders = loaders

    def match(self, file_path: str) -> bool:
        return True

    def load(self, file_path: str) -> Any:
        for loader in self.loaders:
            if loader.match(file_path):
                return loader.load(file_path)

        raise ValueError("Unsupported file path: {}".format(file_path))


# Payload Generator


class PayloadGenerator:
    def __init__(self, file_prefix: str = "file://", images_as_base64=False):
        self.loader = GenericLoader()

        # default loaders
        loaders = [
            FolderLoader(self.loader.load),
            StringLoader(".string"),
            FloatLoader(".float", ".number"),
            IntLoader(".int", ".integer"),
            JsonLoader(".json", ".geojson"),
        ]

        # base64 images parsing
        if images_as_base64:
            loaders.append(Base64ImageLoader(".tif", ".jpg", ".png", ".tiff", ".jpeg"))

        # default loading is local uri file://
        loaders.append(UriLoader(file_prefix))

        self.loader.set_loaders(loaders)

    def generate(self, path: str) -> Any:
        path = os.path.abspath(path)
        print("path={}".format(path))
        result = self.loader.load(path)
        if len(result) == 0:
            return dict()
        return result



import logging
import os
from typing import Tuple, Any

from pesto.common.utils import mkdir
from pesto.ws.core.match_apply import MatchApply, Json

log = logging.getLogger(__name__)


class StringSerializer(MatchApply):
    def __init__(self, job_path: str):
        self.job_path = job_path

    def match(self, schema: Json):
        return schema.get('type') == 'string'

    def convert(self, data: Tuple[Any, str]):
        payload, key = data
        filename = os.path.join(self.job_path, key + '.string')
        mkdir(filename)
        with open(filename, 'w') as outfile:
            outfile.write(str(payload))



import asyncio
import importlib
import logging
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any, Callable

from pesto.ws.core.processing_error import ProcessingError

log = logging.getLogger(__name__)


async def async_exec(callback: Callable) -> Any:
    with ThreadPoolExecutor(max_workers=1) as executor:
        return await asyncio.get_event_loop().run_in_executor(executor, callback)


def load_class(path: str) -> Any:
    """
    Load a class at the provided location. Path is a string of the form: path.to.module.class and conform to the python
    import conventions.

    :param path: string pointing to the class to load
    :return: the requested class object
    """
    try:
        log.info('loading class : [{}]'.format(path))
        module_name, class_name = path.rsplit('.', 1)
        mod = importlib.import_module(module_name)
        return getattr(mod, class_name)
    except Exception:
        raise ProcessingError('Class loading error : expecting path.to.module.ClassName, got : {}'.format(path))



import logging
from typing import Dict, Any

from pesto.ws.core.match_apply import MatchPipeline
from pesto.ws.features.converter.array_converter import ArrayConverter
from pesto.ws.features.converter.image.image_roi import ImageROI
from pesto.ws.features.converter.image_converter import ImageConverter
from pesto.ws.features.converter.metadata_converter import MetadataConverter
from pesto.ws.features.converter.object_converter import ObjectConverter
from pesto.ws.features.converter.primitive_converter import PrimitiveConverter
from pesto.ws.core.pesto_feature import PestoFeature

log = logging.getLogger(__name__)


class PayloadConverter(PestoFeature):
    def __init__(self, image_roi: ImageROI, schema: dict):
        self.schema = schema

        self.input_convert = MatchPipeline([
            ImageConverter(image_roi),
            MetadataConverter(),
            ObjectConverter(self._convert),
            ArrayConverter(self._convert),
            PrimitiveConverter()
        ])

    def process(self, payload: dict) -> dict:
        log.info('payload : '.format(str(payload)))
        return {
            key: self._convert(payload.get(key, None), schema)
            for key, schema in self.schema['properties'].items()
        }

    def _convert(self, value: Any, schema: Dict):
        return self.input_convert.convert((value, schema), schema)



import asyncio
import logging
import traceback
import uuid
from typing import Optional, Tuple, Any, Callable

from pesto.ws.core.payload_parser import PayloadParser, PestoConfig
from pesto.ws.core.pesto_feature import PestoFeatures
from pesto.ws.core.utils import load_class, async_exec
from pesto.ws.features.algorithm_wrapper import AlgorithmWrapper
from pesto.ws.features.converter.image.image_roi import ImageROI, DummyImageROI
from pesto.ws.features.payload_converter import PayloadConverter
from pesto.ws.features.payload_debug import PayloadDebug
from pesto.ws.features.response_serializer import ResponseSerializer
from pesto.ws.features.schema_validation import SchemaValidation
from pesto.ws.features.stateful_response import StatefulResponse
from pesto.ws.features.stateless_response import StatelessResponse
from pesto.ws.service.describe import DescribeService
from pesto.ws.service.job_result import ResultType

log = logging.getLogger(__name__)


class ProcessService:
    PROCESS_CLASS_NAME = 'algorithm.process.Process'

    _algorithm: Optional[Callable] = None
    _describe = None

    @staticmethod
    def init():
        if ProcessService._algorithm is not None:
            raise ValueError('Process Service already loaded !')

        try:
            log.info('ProcessService.init() ...')
            ProcessService._algorithm = load_class(ProcessService.PROCESS_CLASS_NAME)()
            if hasattr(ProcessService._algorithm, 'on_start'):
                log.info('ProcessService.on_start() ...')
                ProcessService._algorithm.on_start()
                log.info('ProcessService.on_start() ... Done !')

            log.info('ProcessService.init() ... Done !')

        except:
            traceback.print_exc()
            log.warning('Algorithm {}.on_start() failure !'.format(ProcessService.PROCESS_CLASS_NAME))

    def __init__(self, url_root: str):
        self.url_root = url_root

    @property
    def service_description(self):
        if ProcessService._describe is None:
            ProcessService._describe = DescribeService(self.url_root).compute_describe()
        return ProcessService._describe

    def process(self, payload: dict) -> dict:
        config = PayloadParser.parse(payload)

        image_roi: Optional[ImageROI] = config.get(PestoConfig.roi)  # if no ROI: None
        active_roi: ImageROI = image_roi or DummyImageROI()  # bypass compute crop info and remove margins in pipeline

        job_id = str(uuid.uuid4().time_low)

        is_stateful = self.service_description['asynchronous'] is True
        input_schema = self.service_description['input']
        output_schema = self.service_description['output']

        common_pipeline = filter(None, [
            SchemaValidation(schema=input_schema),
            active_roi.compute_crop_infos(),
            PayloadConverter(image_roi=image_roi, schema=input_schema),
            PayloadDebug(schema=input_schema),
            AlgorithmWrapper(ProcessService._algorithm),
            active_roi.remove_margin(),
            ResponseSerializer(schema=output_schema, job_id=job_id),
        ])

        if is_stateful:
            pipeline = [
                *common_pipeline,
                StatefulResponse(self.url_root, job_id)
            ]
        else:
            pipeline = [
                *common_pipeline,
                StatelessResponse(self.url_root, job_id, output_schema)
            ]

        return PestoFeatures(pipeline).process(payload)

    async def async_process(self, request_payload: dict) -> Tuple[Any, ResultType]:
        return await asyncio.wait_for(
            async_exec(lambda: self.process(request_payload)),
            timeout=None
        )



import base64
import logging
import os
from typing import Optional

import numpy as np
import rasterio
from pesto.ws.features.converter.image.image_driver import ImageDriver
from pesto.ws.features.converter.image.image_roi import ImageROI

log = logging.getLogger(__name__)

# disable rasterio logs
_rasterio_log = rasterio.logging.getLogger()
_rasterio_log.setLevel(rasterio.logging.ERROR)


class Image:
    @staticmethod
    def from_array(array: np.ndarray) -> 'Image':
        img = Image(array)
        return img

    def to_array(self):
        return self.array

    @staticmethod
    def from_bytes(image_bytes: bytes) -> 'Image':
        with rasterio.MemoryFile(image_bytes) as memory_file:
            with memory_file.open() as file:
                array = file.read()
                return Image.from_array(array=array)

    def to_bytes(self) -> bytes:
        driver = ImageDriver.TIFF if self.bands() > 3 else ImageDriver.PNG

        output_profile = dict(
            driver=driver.driver,
            width=self.width(),
            height=self.height(),
            count=self.bands(),
            dtype=self.array.dtype,
        )

        with rasterio.MemoryFile() as memfile:
            with memfile.open(**output_profile) as dst:
                dst.write(self.array)

            buffer = memfile.read()

        return buffer

    @staticmethod
    def from_uri(uri: str, roi: Optional[ImageROI] = None) -> 'Image':
        try:
            log.debug('opening with rasterio: {}'.format(uri))
            with rasterio.open(uri) as _:
                return _load_image(_, roi)
        except:
            streaming_url = '/vsicurl_streaming/' + uri
            log.debug('trying with streaming: {}'.format(streaming_url))
            with rasterio.open(streaming_url) as _:
                return _load_image(_, roi)

    def to_path(self, path: str) -> str:
        driver = ImageDriver.match_path(path) or ImageDriver.TIFF

        path = os.path.splitext(path)[0] + driver.ext
        os.makedirs(os.path.dirname(path), exist_ok=True)
        log.info('save image [{}]: {} dtype={} to {}'.format(driver, self.array.shape, self.array.dtype, path))
        copy = self.array.astype(np.uint8)
        with rasterio.open(path,
                           mode='w',
                           driver=driver.driver,
                           width=self.width(),
                           height=self.height(),
                           count=self.bands(),
                           dtype=copy.dtype) as output:
            output.write(copy)
        return path

    @staticmethod
    def from_base64(string_b64: str) -> 'Image':
        bytes_b64 = string_b64.encode('utf-8')
        raw_bytes = base64.b64decode(bytes_b64)
        return Image.from_bytes(raw_bytes)

    def to_base64(self) -> str:
        raw_bytes = self.to_bytes()
        bytes_b64 = base64.b64encode(raw_bytes)
        string_b64 = bytes_b64.decode('utf-8')
        return string_b64

    def __init__(self, array: np.ndarray = None):
        if array.ndim == 2:
            array = np.expand_dims(array, 0)
        self.array = array
        log.info('new image : {} dtype={}'.format(self.array.shape, self.array.dtype))

    def width(self) -> int:
        return self.array.shape[2]

    def height(self) -> int:
        return self.array.shape[1]

    def bands(self) -> int:
        return self.array.shape[0]


def _load_image(dataset, roi: Optional[ImageROI]):
    if roi is None:
        data = dataset.read()
    else:
        lines, columns = dataset.shape
        y0, yn, ym = roi.lines.first, roi.lines.number, roi.lines.margin
        x0, xn, xm = roi.columns.first, roi.columns.number, roi.columns.margin
        top, bottom = max(0, y0 - ym), min(lines, y0 + yn + ym)
        left, right = max(0, x0 - xm), min(columns, x0 + xn + xm)
        data = dataset.read(window=((top, bottom), (left, right)))
    return Image.from_array(data)



import os

from pesto.ws.service.job_list import JobListService


class JobStatusService:
    def __init__(self, url_root: str, job_id: str):
        self.job_id = job_id
        self.job_path = os.path.join(JobListService.PESTO_WORKSPACE, self.job_id)
        self.url_root = url_root

    def get_status(self) -> dict:
        if not os.path.exists(self.job_path):
            message = 'path does not exists : {}'.format(self.job_path)
            raise RuntimeError(message)
        if not os.listdir(self.job_path):
            return {
                'status': 'RUNNING',
                'progress': 0,
            }
        return {
            'status': 'DONE',
            'progress': 100,
            'link': '{}/api/v1/jobs/{}/results'.format(self.url_root, self.job_id)
        }



from setuptools import setup, find_packages

from pesto import version

entry_points = {
    'console_scripts': [
        'pesto=pesto.cli.app:main',
        'processing=pesto.ws.app:main',
    ],
}

setup(
    name='processing-factory',
    version=version.PESTO_VERSION,
    packages=find_packages(),
    maintainer='Airbus Defense & Space',
    description='Geo Processing Micro Service SDK',
    install_requires=[line for line in open('requirements.txt')],
    entry_points=entry_points,
    include_package_data=True,
)



import asyncio
import logging
from typing import Any

from sanic import Blueprint, response
from sanic.exceptions import ServerError
from sanic.request import Request
from sanic.response import HTTPResponse
from sanic_openapi import doc

from pesto.ws.service.describe import DescribeService
from pesto.ws.service.job_delete import JobDeleteService
from pesto.ws.service.job_list import JobListService
from pesto.ws.service.job_result import JobResultService, ResultType
from pesto.ws.service.job_status import JobStatusService
from pesto.ws.service.process import ProcessService

log = logging.getLogger('pesto')
v1 = Blueprint('v1', url_prefix='/api/v1')

processing_semaphore = None


@v1.listener('before_server_start')
async def init(sanic, loop):
    global processing_semaphore
    processing_semaphore = asyncio.Semaphore(1, loop=loop)
    ProcessService.init()


@v1.route('/openapi')
@doc.summary('Open API specification of this service in YAML format')
async def openapi(request: Request) -> HTTPResponse:
    return await response.file('/etc/pesto/api_geo_process_v1.0.yaml')


@v1.route('/describe')
@doc.summary('Describes processing service')
@doc.produces('JSON')
async def describe(request: Request) -> HTTPResponse:
    try:
        url_root = _get_url_root(request)
        service_description = DescribeService(url_root).compute_describe()
        return response.json(service_description)
    except Exception as e:
        log.exception(e)
        raise ServerError('service.json does not exist or is not json serializable',
                          status_code=500)


@v1.route('/version')
@doc.summary('Processing version')
def version_get(request: Request) -> HTTPResponse:
    try:
        version = DescribeService.compute_version()
        return response.json(version)
    except Exception as e:
        log.exception(e)
        message = 'Could not load version file : {}'.format(DescribeService.VERSION_CONTENT_PATH)
        raise ServerError(message, status_code=500)


@v1.route('/health')
@doc.summary('Check if service is alive')
def get(request: Request) -> HTTPResponse:
    result = response.text('OK', 200)
    result.headers['content-type'] = 'text/plain'
    return result


@v1.route('/jobs', methods=['POST'])
@doc.summary('Launch process execution')
def jobs_post(request: Request) -> HTTPResponse:
    return response.redirect(request.app.url_for('api.process'))


@v1.route('/jobs', methods=['GET'])
@doc.summary('List jobs')
def jobs_get(request: Request) -> HTTPResponse:
    url_root = _get_url_root(request)
    result = JobListService().job_list(url_root)
    return response.json(result)


@v1.route('/jobs/<job_id>')
@doc.summary('Retrieve execution status')
def jobs_id_delete(request: Request, job_id: str) -> HTTPResponse:
    # Job processing is always synchronous (so no cancelling is possible)
    pass


@v1.route('/jobs/<job_id>/status')
@doc.summary('Retrieve execution status')
def jobs_id_status_get(request: Request, job_id: str) -> HTTPResponse:
    url_root = _get_url_root(request)
    status = JobStatusService(url_root, job_id).get_status()
    return response.json(status)


@v1.route('/jobs/<job_id>/results')
@doc.summary('Retrieve execution status')
def jobs_id_results_get(request: Request, job_id: str) -> HTTPResponse:
    url_root = _get_url_root(request)
    results = JobResultService(url_root, job_id).get_results()
    return response.json(results)


# @v1.route('/jobs/<job_id>/results/<result_id>', methods=['DELETE'])
@doc.summary('Retrieve execution status')
def jobs_results_delete(request: Request, job_id: str) -> HTTPResponse:
    try:
        url_root = _get_url_root(request)
        JobDeleteService(url_root, job_id).delete()
        return HTTPResponse()
    except Exception as e:
        log.exception(e)
        message = 'Error while deleting : job_id = {}'.format(job_id)
        raise ServerError(message, status_code=500)


@v1.route('/jobs/<job_id>/results/<result_id>', methods=['GET'])
@doc.summary('Retrieve execution status')
async def jobs_results_id_get(request: Request, job_id: str, result_id: str) -> HTTPResponse:
    try:
        url_root = _get_url_root(request)
        output, data_type = JobResultService(url_root, job_id).get_partial_result(result_id)
        return await _prepare_response(output, data_type)
    except Exception as e:
        log.exception(e)
        message = 'Error while retrieving results : job_id = {}, result_id = {}'.format(job_id, result_id)
        raise ServerError(message, status_code=500)


@v1.route('/jobs/<job_id>/results/<result_id>', methods=['DELETE'])
@doc.summary('Retrieve execution status')
def jobs_results_id_delete(request: Request, job_id: str, result_id: str) -> HTTPResponse:
    try:
        url_root = _get_url_root(request)
        JobDeleteService(url_root, job_id).delete_partial(result_id)
        return HTTPResponse({})
    except Exception as e:
        log.exception(e)
        message = 'Error while deleting : job_id = {}, result_id = {}'.format(job_id, result_id)
        raise ServerError(message, status_code=500)


# Implementation specifics endpoints
@v1.route('/process', methods=['POST'])
@doc.summary('Launch process execution')
async def process_post(request: Request) -> HTTPResponse:
    url_root = _get_url_root(request)
    request_payload = request.json
    if request_payload is None:
        raise ServerError('json payload is empty, or payload is not json', status_code=500)

    if processing_semaphore.locked():
        raise ServerError('a processing is already running', status_code=429)

    await processing_semaphore.acquire()
    try:
        future = ProcessService(url_root).async_process(request_payload)
        output, data_type = await asyncio.wait_for(future, timeout=None)
        return await _prepare_response(output, data_type)
    except Exception as e:
        log.exception(e)
        message = 'Error while processing, see logs for more details : {}'.format(str(e))
        raise ServerError(message, status_code=500)
    finally:
        processing_semaphore.release()


def _get_url_root(request: Request) -> str:
    return '{0.scheme}://{0.host}'.format(request)


async def _prepare_response(output: Any, data_type: ResultType):
    if data_type == ResultType.json:
        return response.json(output)
    elif data_type == ResultType.file:
        return response.text(output)
    elif data_type == ResultType.image:
        with open(output, "rb") as f:
            image = f.read()
        return response.raw(body=image, headers={"Content-Type": "image/png"})



import os

from pesto.cli.core.utils import PESTO_LOG


def list_builds(workspace_root):
    PESTO_LOG.info('Processing Factory repository path :'.format(workspace_root))
    PESTO_LOG.info('list of available builds :')

    for name in os.listdir(workspace_root):
        if name.startswith('.'):
            continue
        for version in os.listdir(os.path.join(workspace_root, name)):
            id = '{}:{}'.format(name, version)
            PESTO_LOG.info(''' {0} :
            pesto build {0}
             '''.format(id))



from typing import Any, Tuple

from pesto.ws.core.match_apply import MatchApply, Json


class MetadataConverter(MatchApply):
    def match(self, schema: Json):
        ref = schema.get('$ref')
        return ref in [
            '#/definitions/Metadata',
            '#/definitions/Metadatas',
        ]

    def convert(self, data: Tuple[Any, Json]):
        payload, schema = data
        return payload



from sanic import Sanic
from sanic_openapi import swagger_blueprint, openapi_blueprint

from pesto.ws.v1 import v1

import logging

_logging_format = '[%(asctime)s] %(process)d-%(levelname)s '
_logging_format += '%(module)s::%(funcName)s():l%(lineno)d: '
_logging_format += '%(message)s'

#logging.basicConfig(format=_logging_format, level=logging.DEBUG)
logging.basicConfig(format=_logging_format)

# Declare Sanic application
app = Sanic(__name__)

# API Configuration
app.blueprint(openapi_blueprint)
app.blueprint(swagger_blueprint)
app.blueprint(v1)

app.config.API_VERSION = '1.0.0'
app.config.API_TITLE = 'GeoProcessing SDK API'
app.config.API_DESCRIPTION = 'GeoProcessing SDK API'
app.config.API_TERMS_OF_SERVICE = 'Apache 2.0'
app.config.API_PRODUCES_CONTENT_TYPES = ['application/json', 'mimetypes/jpeg', 'mimetypes/tiff']
app.config.API_CONTACT_EMAIL = 'tbd@airbus.com'

app.config.KEEP_ALIVE = False

TIMEOUT = 1_000_000
app.config.REQUEST_TIMEOUT = TIMEOUT
app.config.RESPONSE_TIMEOUT = TIMEOUT
app.config.KEEP_ALIVE_TIMEOUT = TIMEOUT


def main():
    app.run(host="0.0.0.0", port=8080, debug=False)


if __name__ == '__main__':
    main()



import logging
import os

_logging_format = '[%(asctime)s] %(process)d-%(levelname)s '
_logging_format += '%(module)s::%(funcName)s():l%(lineno)d:\n'
_logging_format += '%(message)s'

logging.basicConfig(format=_logging_format, level=os.environ.get("PESTO_LOGLEVEL") or "INFO")

PESTO_LOG = logging.getLogger("pesto")



import os
from pathlib import Path

PESTO_WORKSPACE = os.environ.get('PESTO_WORKSPACE', '{}/.pesto'.format(Path.home()))
_PESTO_PROFILE = os.environ.get('PESTO_PROFILE', '')


class Pesto:
    @staticmethod
    def get_profile():
        return _PESTO_PROFILE

    @staticmethod
    def is_profile_active(profile: str) -> bool:
        return profile in _PESTO_PROFILE



from pesto.cli.core.build_config import BuildConfig
from pesto.cli.core.docker_builder import DockerBuilder

def test_docker_image_name():
    # given
    config = BuildConfig(
        name='my-service',
        version='1.2.3',
        profiles=['p1', 'p2']
    )

    # when
    actual = config.docker_image_name

    # then
    expected = '{}:{}-{}'.format(config.name, config.version, '-'.join(config.profiles))
    assert actual == expected


def test_network():
    config = BuildConfig(
        name='my-service',
        version='1.2.3',
        profiles=['p1', 'p2'],
        network='test_network')
    assert config.network == "test_network"


def test_pythonpath():
    # given
    build_config = BuildConfig(
        name='my-service',
        version='1.2.3',
        profiles=['p1', 'p2'],
        workspace=".")
    requirements={"dockerBaseImage":{}, "requirements":{}, "environments":{}}

    dockerbuilder = DockerBuilder(requirements, build_config).dockerfile()

    # when
    actual = dockerbuilder.split("\n")[17]
    dockerfile_lines = dockerbuilder.split("\n")
    for line in dockerfile_lines:
        if line[:14] == "ENV PYTHONPATH":
            actual = line

    # then
    expected = 'ENV PYTHONPATH=$PYTHONPATH${PYTHONPATH:+:}/opt/my-service'
    assert actual == expected



import logging
import os
from enum import Enum

log = logging.getLogger(__name__)


class ImageDriver(Enum):
    TIFF = ('.tif', 'GTiff')
    PNG = ('.png', 'PNG')
    JPG = ('.jpg', 'JPEG')

    @staticmethod
    def match_path(path: str):
        _, ext = os.path.splitext(path)
        ext = ext.replace('.', '')

        if ext in ['jpeg', 'jpg']:
            return ImageDriver.JPG
        elif ext in ['tif', 'tiff']:
            return ImageDriver.TIFF
        elif ext in ['png']:
            return ImageDriver.PNG
        else:
            log.error('image : unsupported extension : {}'.format(ext))
            return None

    @property
    def ext(self) -> str:
        return self.value[0]

    @property
    def driver(self) -> str:
        return self.value[1]



import shlex
import subprocess
from typing import List

from pesto.cli.core.utils import PESTO_LOG
from pesto.cli.core.build_config import BuildConfig


def push(build_config_path, registry_path, profiles: List[str]):
    PESTO_LOG.info('Push docker image to registry : {}'.format(registry_path))

    build_config = BuildConfig.from_path(path=build_config_path, profiles=profiles)
    PESTO_LOG.info('build configuration : {}'.format(build_config))

    algo_name = build_config.name
    algo_version = build_config.version
    local_image_name = '{}:{}'.format(algo_name, algo_version)
    remote_image_name = '{}/{}'.format(registry_path, local_image_name)

    print(local_image_name)
    print(remote_image_name)

    PESTO_LOG.info('tag image {} as {}'.format(local_image_name, remote_image_name))
    subprocess.call(shlex.split('docker tag {} {}'.format(local_image_name, remote_image_name)))
    PESTO_LOG.info('push to gcloud : {}'.format(remote_image_name))
    subprocess.call(shlex.split('gcloud docker -- push  {}'.format(remote_image_name)))



from enum import Enum


class PestoFiles(Enum):
    description = 'api/description.json'
    input_schema = 'api/input_schema.json'
    output_schema = 'api/output_schema.json'
    config_schema = 'api/config_schema.json'
    config = 'api/config.json'
    version = 'api/version.json'
    service = 'api/service.json'

    build = 'build/build.json'
    requirements = 'build/requirements.json'

    @staticmethod
    def required():
        return set(PestoFiles).difference({
            PestoFiles.service,
            PestoFiles.build
        })
